#!/bin/python3
#%%
from time import localtime, time, sleep, asctime
# t_init = time()

#%%
import base64, datetime, functools, gzip, io, importlib, json, os, pickle, re, sys, tarfile, uuid
from shutil import copy2, copytree
from collections import Counter
from hashlib import sha256, md5
from itertools import combinations, product
from random import shuffle, random, seed
from multiprocessing import cpu_count, Pool

#%%
import numpy as np
import pandas as pd

#%%
from tools.std_lib.basic import count_down, ls_dirs, ls_dirs_r, ls_files, ls_files_r, replace_slash
from tools.data_processing.basic_data_tools import colorize_str, get_file_size, human_readable_size, pickle_load,  pickle_dump, get_timestamp, check_dataframe, split_dataframe, stat_month_df

#%%
from tools.std_lib.log import ColorfulLogger, apply_color
C = apply_color
log = ColorfulLogger(basename='easy_pandas.py')
t_print = lambda *args, **kwargs : log(*args, **kwargs)

#%%
# sys

def import_py_module(file_path: str, module_name = None):
    log.info(f'file_path =', C(file_path, color='underline,blue'))
    if not os.path.exists(file_path):
        log.error('not exists')
        raise ValueError(f'not exists')
    if module_name is None:
        module_name = file_path
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    log.info('dir(module) =', sorted(dir(module)))
    return module


#%%
# math

# def get_int_from_str(s):
#     if 'str' in s:
#         s = s.str
#     findall = re.findall('\d+', s)
#     if findall:
#         r = int(findall[0])
#     else:
#         r = float('nan')
#     return r


#%%
# bash

# replace_slash = lambda p: p.replace('\\', '/')


# def add_prefix_to_filename(file_path, prefix, sep = '-'):
#     base_name = os.path.basename(file_path)
#     new_file_name = prefix + sep + base_name
#     new_path = os.path.join(os.path.dirname(file_path), new_file_name)
#     print(f'new_path = \'{new_path}\'')
#     return new_path


# def add_suffix_to_filename(path: str, suffix = '', var = None, sep = '-'):
#     p1, p2 = os.path.splitext(path)
#     if len(suffix) > 0:
#         suffix = sep + str(suffix)
#     if var is not None:
#         if isinstance(var, pd.DataFrame):
#             suffix += f'{sep}shape_%d_%d'%var.shape
#         elif isinstance(var, dict) or isinstance(var, list):
#             suffix += f'{sep}n{len(var)}'
#     new_path = p1 + suffix + p2
#     print(f'new_path = \'{new_path}\'')
#     return new_path


# def all_in_dir(dir_path, ls = []):
#     ls_dir = os.listdir(dir_path)
#     if ls:
#         if type(ls) == list:
#             if len(set(ls) - set(ls_dir)) == 0:
#                 r = True
#             else:
#                 r = False
#         elif type(ls) == str:
#             pattern = ls
#             r = True
#             for p in ls_dir:
#                 if not re.findall(pattern, p):
#                     r = False
#                     break
#         else:
#             print('ERROR: all_in_dir() type error')
#     else:
#         r = True
#     return r


# def any_in_dir(dir_path, ls = []):
#     ls_dir = os.listdir(dir_path)
#     if ls:
#         if type(ls) == list:
#             if len(set(ls) - set(ls_dir)) < len(ls):
#                 r = True
#             else:
#                 r = False
#         elif type(ls) == str:
#             pattern = ls
#             r = False
#             for p in ls_dir:
#                 if re.findall(pattern, p):
#                     r = True
#                     break
#         else:
#             print('ERROR: any_in_dir(): type error')
#     else:
#         r = True
#     return r


# def ls_in_dir(dir_path, all_ls = [], any_ls = []):
#     r1 = all_in_dir(dir_path, all_ls)
#     r2 = any_in_dir(dir_path, any_ls)
#     r = r1 & r2
#     return r


# def ls_files_r(path = './', slash = False, pattern = '', show = False):
#     if os.path.exists(path):
#         r = []
#         for k1, _, k2 in os.walk(path):
#             for k22 in k2:
#                 if re.findall(pattern, k22, re.IGNORECASE):
#                     p = os.path.join(k1, k22)
#                     r.append(p)
#         r = sorted(r)
#         if slash:
#             r = list(map(replace_slash, r))
#         if show:
#             [print(f"path = '{k}'") for k in r]
#     else:
#         print(f'ERROR: {path} does not exist.')
#     return r


# def ls_dirs_r(path = './', slash = False, all_ls = [], any_ls = [], pattern = '', show = False):
#     if os.path.exists(path):
#         r = [path] if re.findall(pattern, path) else []
#         for k1, k2, _ in os.walk(path):
#             # r.append(k1)
#             for k22 in k2:
#                 if re.findall(pattern, k22, re.IGNORECASE):
#                     p = os.path.join(k1, k22)
#                     r.append(p)
#         r = sorted(r)
#         if slash:
#             r = list(map(replace_slash, r))
#         r = [p for p in r if ls_in_dir(p, all_ls, any_ls)]
#         if show:
#             [print(f"path = '{k}'") for k in r]
#     else:
#         print(f'ERROR: {path} does not exist.')
#     return r


# def ls_files(path, slash = False, pattern = '', show = False):
#     ps = sorted(os.listdir(path))
#     paths = [os.path.join(path, k) for k in ps if re.findall(pattern, k)]
#     dirs = []
#     for p in paths:
#         if os.path.isfile(p):
#             dirs.append(p)
#     if slash:
#         dirs = [replace_slash(p) for p in dirs]
#     if show:
#         [print(f"path = '{k}'") for k in dirs]
#     return dirs


# def ls_dirs(path, slash = False, all_ls = [], any_ls = [], pattern = '', show = False):
#     ps = sorted(os.listdir(path))
#     paths = [os.path.join(path, k) for k in ps if re.findall(pattern, k)]
#     dirs = []
#     for p in paths:
#         if os.path.isdir(p):
#             dirs.append(p)
#     if slash:
#         dirs = [replace_slash(p) for p in dirs]
#     dirs = [p for p in dirs if ls_in_dir(p, all_ls, any_ls)]
#     if show:
#         [print(f"path = '{k}'") for k in dirs]
#     return dirs


# def find_latest_file(path: str, pattern: str):
#     latest_timestamp = 0
#     latest_file = None
#     for file in os.listdir(path):
#         if re.findall(pattern, file):
#             full_path = os.path.join(path, file)
#             timestamp = os.path.getmtime(full_path)
#             if timestamp > latest_timestamp:
#                 latest_timestamp = timestamp
#                 latest_file = full_path
#     print(f'latest_file = {latest_file}')
#     return latest_file


# def extend_path(path, mode=None, n = 0):
#     ps = sorted(os.listdir(path))
#     paths = [os.path.join(path, k) for k in ps]
#     if mode == None:
#         r = paths[n]
#     elif mode == 'f':
#         files = []
#         for p in paths:
#             if os.path.isfile(p):
#                 files.append(p)
#             r = files[n]
#     elif mode == 'd':
#         dirs = []
#         for p in paths:
#             if os.path.isdir(p):
#                 dirs.append(p)
#         r = dirs[n]
#     return r


# def human_readable_size(file_size: int):
#     size = file_size
#     unit_index = 0
#     units = ['B', 'KiB', 'MiB', 'GiB', 'TiB']
#     while size >= 1024 and unit_index < len(units) - 1:
#         size /= 1024.0
#         unit_index += 1
#     ret = f'{size:.2f} {units[unit_index]}'
#     return ret


# def get_file_size(file_path, verbose = False, eta = None):
#     if isinstance(file_path, bytes):
#         file_size = len(file_path)
#         file_path = colorize_str('bytes', color=[4, 94])
#     elif isinstance(file_path, str):
#         file_size = os.path.getsize(file_path)
#         file_path = colorize_str(file_path, color=[4, 94])
#     if verbose or eta is not None:
#         print(f"File size: \n{human_readable_size(file_size=file_size)}    {file_path}")
#         if eta is not None:
#             if eta != 0:
#                 v = human_readable_size(file_size / eta)
#             else:
#                 v = 'inf'
#             print(f'rate = {v}/s, eta = {eta:.2f}s')
#     return file_size


def copy_to_dir(src_path: str, dst_dir_path: str):
    dst_path = os.path.join(dst_dir_path, os.path.basename(src_path))
    if not os.path.exists(src_path):
        raise ValueError(f'src_path = \'{src_path}\' does not exist.')
    if os.path.exists(dst_path):
        raise ValueError(f'dst_path = \'{dst_path}\' exists.')
    while src_path[-1] in '/\\':
        src_path = src_path[:-1]
    if os.path.isfile(dst_dir_path):
        raise ValueError(f'dst_dir_path = \'{dst_dir_path}\' is a file.')
    elif not os.path.exists(dst_dir_path):
        os.makedirs(dst_dir_path)
    if os.path.isfile(src_path):
        copy2(src_path, dst_path)
    elif os.path.isdir(src_path):
        copytree(src_path, dst_path, dirs_exist_ok=True)
    print(f'copied {src_path} to {dst_path}')
    return None

#%%

def read_txt(path: str, to_list = False, encoding = 'utf-8'):
    with open(path, 'r', encoding=encoding) as f:
        s = f.read()
    if to_list:
        s = [k.strip() for k in s.split('\n') if k]
    return s


def save_txt(string: str, path: str, encoding = 'utf-8'):
    with open(path, 'w', encoding=encoding) as f:
        f.write(string)
    return None


#%%
# time

def date_str(fw = 0):
    dt = asctime(localtime(time()+fw))
    return dt


# def t_print(s):
#     if type(s) == dict:
#         try:
#             s = '\n' + json_dumps(s)
#         except:
#             s = str(s)
#     elif type(s) != str:
#         s = '\n' + str(s)
#     print(f'T={time() - t_init:1.0f}s {s}')
#     return None

# def get_color_prefix(n: object):
#     if isinstance(n, int):
#         prefix = f'\033[{n}m'
#     elif isinstance(n, list):
#         prefix = ''.join([f'\033[{k}m' for k in n])
#     return prefix

# def colorize_str(string: str, color = [91], show = False):
#     string = str(string)
#     if isinstance(color, int):
#         color = [color]
#     colorful_string = get_color_prefix(color) + string + '\033[0m'
#     if show:
#         print(colorful_string)
#     return colorful_string


# def t_print(*args, color = None):
#     text = f'T={time() - t_init:1.0f}s'
#     for var in args:
#         if isinstance(var, pd.Series) or isinstance(var, pd.DataFrame):
#             if text[-1] == ' ':
#                 text = text[:-1] + '\n'
#             elif text[-1] != '\n':
#                 text = text + '\n'
#             end = '\n'
#         elif isinstance(var, list) or isinstance(var, dict) :
#             if text[-1] == ' ':
#                 text = text[:-1] + '\n'
#             elif text[-1] != '\n':
#                 text = text + '\n'
#             end = '\n'
#             var = json_dumps(dict_to_json(var))
#             var = colorize_str(var, 92)
#         elif isinstance(var, set) :
#             if text[-1] == ' ':
#                 text = text[:-1] + '\n'
#             elif text[-1] != '\n':
#                 text = text + '\n'
#             end = '\n'
#         else:
#             if text[-1] != '\n':
#                 text = text + ' '
#             end = ''
#         text += str(var) + end
#     while text[-1] in ['\n', ' ']:
#         text = text[:-1]
#     if color is not None:
#         text = colorize_str(text, color=color)
#     print(text)
#     return None


def parse_date(s):
    r = s
    if type(s) == str:
        s = s.strip()
        if len(re.findall('-', s)) == 2 and len(s) <= 10:
            r = datetime.datetime.strptime(s, '%Y-%m-%d')
        elif len(re.findall('/', s)) == 2 and len(s) <= 10:
            r = datetime.datetime.strptime(s, '%Y/%m/%d')
        elif len(s) == 8:
            r = datetime.datetime.strptime(s, '%Y%m%d')
        else:
            r = datetime.datetime.strptime(s, '%Y-%m-%d %H:%M:%S')
    elif type(s) == int:
        r = parse_date(str(s))
    elif type(s) == float and not np.isnan(s) and s == int(s):
        r = parse_date(str(int(s)))
    return r


# def count_down(n: int):
#     if type(n) == list:
#         s = sorted(set(n))
#         n = (s[1] - s[0]) * random() + s[0] + 1
#     n = int(np.floor(n))
#     for k in range(n):
#         print(f'{n - k - 1}s  \r', end = '')
#         sleep(1)
#     return None


def get_year_month_ls():
    s = []
    for year in range(2015, 2025+1):
        for month in range(1, 12+1):
            n = 100*year + month
            s.append(n)
    return s

year_month_ls = get_year_month_ls()

count_by_month = lambda date_s: date_s.apply(lambda d: datetime.datetime(d.year, d.month, 1)).value_counts().sort_index()


# get_timestamp = lambda : strftime("%Y%m%d%a%H%M%S", localtime())

# def get_timestamp(dt: datetime.datetime = None):
#     time_format = "%Y%m%d%a%H%M%S"
#     if dt is None:
#         result = strftime(time_format, localtime())
#     elif isinstance(dt, datetime.datetime):
#         result = dt.strftime(time_format)
#     return result



#%%

get_uuid_hex = lambda: uuid.uuid4().hex


def get_bytes_hash(bs: bytes, alg = 'sha256'):
    if alg == 'sha256':
        m = sha256()
    elif alg == 'md5':
        m = md5()
    else:
        t_print(f'alg = {alg} not found')
    m.update(bs)
    r = m.hexdigest()
    return r


def get_file_hash(path: str, alg = 'sha256'):
    with open(path, 'rb') as f:
        r = get_bytes_hash(f.read(), alg=alg)
    return r


def dump_files_hash(path, dump = True, timestamp = None):
    t_print('dir_path = ', path)
    files = sorted(os.listdir(path))
    file_paths = [os.path.join(path, f) for f in files if os.path.isfile(os.path.join(path, f))]
    file_hash_dict = {}
    for k, file_path in enumerate(file_paths):
        f = os.path.basename(file_path)
        t_print(f'{k+1}/{len(file_paths)}', f)
        file_hash_dict[f] = {}
        file_hash_dict[f]['size'] = os.path.getsize(file_path)
        file_hash_dict[f]['sha256'] = get_file_hash(file_path, alg='sha256')
        file_hash_dict[f]['md5'] = get_file_hash(file_path, alg='md5')
    if dump:
        if timestamp is None:
            timestamp = get_timestamp()
        json_name = f'file_hash_{timestamp}.json'
        json_path = os.path.join(path, json_name)
        json_dump(file_hash_dict, json_path)
    t_print('end of dump_files_hash()')
    return file_hash_dict


def check_file_hash(path: str):
    if os.path.isdir(path):
        dir_path = path
        path = ls_files(dir_path, pattern=r'^file_hash_\d{8}\w{3}\d{6}.json$')[-1]
    dir_path = os.path.dirname(path)
    print(f'file_hash_path = {path}')
    info = json_load(path=path)
    file_status = {}
    for file_name, file_info in info.items():
        file_path = os.path.join(dir_path, file_name)
        file_status[file_name] = float('nan')
        if os.path.isfile(file_path):
            file_status[file_name] = False
            old_size = file_info['size']
            new_size = os.path.getsize(file_path)
            if old_size == new_size  :
                old_sha256_str = file_info['sha256']
                new_sha256_str = get_file_hash(file_path)
                file_status[file_name] = old_sha256_str == new_sha256_str
    s = pd.Series(file_status)
    print(s)
    print('\nvalue_counts:')
    t_print(s.fillna('NaN').value_counts().to_dict())
    return file_status


def get_hash(s = b'', n = None, alg = 'sha256', salt = ''):
    if type(s) == bytes:
        r = get_bytes_hash(s, alg=alg)
    elif type(s) == str:
        s = s + salt
        s = s.encode('utf-8')
        r = get_hash(s, alg=alg, salt=salt)
    elif type(s) == list or type(s) == dict:
        s = json_dumps(dict_to_json(s))
        r = get_hash(s, alg=alg, salt=salt)
    # elif type(s) == pd.core.indexes.numeric.Int64Index:
    #     s = s.to_list()
    #     r = get_hash(s, alg=alg, salt=salt)
    elif type(s) in [float, int]:
        s = str(s)
        r = get_hash(s, alg=alg, salt=salt)
    elif isinstance(s, pd.Series):
        s = s.to_dict()
        r = get_hash(s, alg=alg, salt=salt)
    elif isinstance(s, pd.DataFrame):
        df = s
        d = df.to_dict()
        r = get_hash(d, alg=alg, salt=salt)
    if n is not None:
        r = r[:n]
    return r


def get_bytes_sha256(bs: bytes):
    m = sha256()
    m.update(bs)
    r = m.hexdigest()
    return r


def get_file_sha256(path: str):
    with open(path, 'rb') as f:
        r = get_bytes_sha256(f.read())
    return r


def get_files_sha256(path: str):
    path = path.rstrip('/\\')
    files = ls_files_r(path, slash=True)
    d = {f[len(path)+1:]: get_file_sha256(f) for f in files}
    return d


def get_sha256(s, n = 0):
    if type(s) == bytes:
        r = get_bytes_sha256(s)
    elif type(s) == str:
        s = s.encode('utf-8')
        r = get_sha256(s)
    elif type(s) == list or type(s) == dict:
        s = json_dumps(dict_to_json(s))
        r = get_sha256(s)
    elif type(s) == pd.core.indexes.numeric.Int64Index:
        s = s.to_list()
        r = get_sha256(s)
    if n > 0:
        r = r[:n]
    return r


#%%
# base64

def pack(input_path: str, output_path = None, alg = 'gz'):
    if input_path[-1] in '/\\':
        input_path = input_path[:-1]
    if output_path is None:
        output_path = f'{input_path}.tar.{alg}'
        if os.path.exists(output_path):
            print(f'warning: output_path = \'{output_path}\' exists.')
    with tarfile.open(output_path, f'w:{alg}') as tar:
        arcname = os.path.basename(input_path)
        tar.add(input_path, arcname=f'{arcname}')
    return output_path


def unpack(input_path, output_path = None):
    if output_path is None:
        output_path = os.path.dirname(input_path)
    with tarfile.open(input_path, 'r') as tar:
        tar.extractall(path=output_path)
        files = tar.getnames()
    paths = [os.path.join(output_path, p) for p in files]
    if '\\' not in output_path or '/' in output_path:
        paths = [k.replace('\\', '/') for k in paths]
    return paths


bytes_to_base64 = lambda b: base64.b64encode(b).decode('utf-8')


def file_to_base64(path: str, c = 100, output_path = None, show = False):
    with open(path, 'rb') as f:
        b = f.read()
    base64_str = bytes_to_base64(b)
    if c is not None:
        lines = []
        for k in range(0, len(base64_str), c):
            lines.append(base64_str[k: k+c])
        base64_str = '\n'.join(lines)
        base64_str += '\n'
    if show:
        print(base64_str)
    if output_path is not None:
        if output_path == '':
            output_path = path + '.base64.txt'
        with open(output_path, 'w', encoding='utf-8') as f:
            f.write(base64_str)
        print(f'base64.txt saved to \'{output_path}\'')
    return output_path


def base64str_to_file(base64_str, path):
    binary_data = base64.b64decode(base64_str)
    if os.path.exists(path):
        print(f'warning: path = \'{path}\' exists.')
    with open(path, 'wb') as f:
        f.write(binary_data)
    return None


def compress_and_encode_str(s, tag = False):
    b = s.encode('utf-8')
    compressed = gzip.compress(b)
    encoded = base64.b64encode(compressed).decode('utf-8')
    if tag:
        encoded = f'<gzip_base64>{encoded}</gzip_base64>'
    return encoded


def decode_and_decompress(encoded, tag = 'auto', parse_json = True):
    if tag == 'auto':
        findall_ls = re.findall(r'<gzip_base64>(.*?)</gzip_base64>', encoded)
        if len(findall_ls) == 1:
            encoded = findall_ls[0]
    compressed = base64.b64decode(encoded)
    decompressed = gzip.decompress(compressed)
    s = decompressed.decode('utf-8')
    if parse_json:
        s = json_loads(s)
    return s


def xor_binary(binary):
    ret = bytes([a ^ b for a, b in zip(binary[0], binary[1])])
    return ret


def cut_long_str(s: str, length = 2**20):
    ret = [
        s[n:n+length] for n in range(0, len(s), length)
        ]
    return ret



def multiprocessing_xor_binary(binary, length = 2**20):
    a_bytes_ls = cut_long_str(binary[0], length=length)
    b_bytes_ls = cut_long_str(binary[1], length=length)
    if len(a_bytes_ls) == len(b_bytes_ls):
        length_str = human_readable_size(length)
        t_print(f'cut long str/bytes into {length_str}*{len(a_bytes_ls)}')
    else:
        raise ValueError
    binary_ls = list(zip(a_bytes_ls, b_bytes_ls))
    bytes_ls = multiprocessing_apply(xor_binary, binary_ls)
    ret = bytes(b''.join(bytes_ls))
    return ret
    

# def multi_processing_xor(a_bytes: bytes, b_bytes: bytes):
#     t_print('multi_processing_xor()')
#     binary_list = list(zip(a_bytes, b_bytes))
#     ret = multiprocessing_apply(xor, binary_list)
#     ret = bytes(ret)
#     return ret



def xor_encrypt(data, key, format = 'b64', length = 2**20):
    if isinstance(data, bytes):
        data_bytes = data
    elif isinstance(data, str):
        data_bytes = data.encode('utf-8')
    key_bytes = key.encode('utf-8')
    extended_key = (key_bytes * (len(data_bytes) // len(key_bytes) + 1))[:len(data_bytes)]
    t0 = time()
    # encrypted_bytes = bytes([b ^ k for b, k in zip(data_bytes, extended_key)])
    # encrypted_bytes = xor_binary((data_bytes, extended_key))
    encrypted_bytes = multiprocessing_xor_binary((data_bytes, extended_key), length=length)
    t1 = time()
    get_file_size(encrypted_bytes, eta=t1-t0)
    if format == 'hex':
        encrypted_data = encrypted_bytes.hex()
    elif format in ['base64', 'b64']:
        encrypted_data = base64.b64encode(encrypted_bytes).decode('utf-8')
    elif format in ['bytes', 'binary']:
        encrypted_data = encrypted_bytes
    print(f'encrypt_length: {len(data)} -> {len(encrypted_data)}')
    return encrypted_data


def xor_decrypt(data, key, input_format = 'b64', output_format = 'binary', length = 2**20):
    if isinstance(data, str) and input_format == 'hex':
        data_bytes = bytes.fromhex(data)
    elif isinstance(data, str) and input_format in ['base64', 'b64']:
        data_bytes = base64.b64decode(data)
    elif isinstance(data, bytes) :
        data_bytes = data
    key_bytes = key.encode('utf-8')
    extended_key = (key_bytes * (len(data_bytes) // len(key_bytes) + 1))[:len(data_bytes)]
    t0 = time()
    # decrypted_bytes = bytes([b ^ k for b, k in zip(data_bytes, extended_key)])
    # decrypted_bytes = xor_binary((data_bytes, extended_key))
    decrypted_bytes = multiprocessing_xor_binary((data_bytes, extended_key), length=length)
    t1 = time()
    get_file_size(decrypted_bytes, eta=t1-t0)
    if output_format in ['bytes', 'binary']:
        decrypted_data = decrypted_bytes
    elif output_format in ['str', 'string']:
        decrypted_data = decrypted_bytes.decode('utf-8')
    print(f'decrypt_length: {len(data)} -> {len(decrypted_data)}')
    return decrypted_data


def encrypte_and_pickle_dump(data: bytes, path: str, key: str):
    t_print('pickle.dumps()')
    t0 = time()
    binary = pickle.dumps(data)
    t1 = time()
    get_file_size(binary, eta = t1 - t0)
    t_print('xor_encrypt()')
    encrypted_binary = xor_encrypt(binary, key=key, format='binary')
    t_print('f.write()')
    t2 = time()
    with open(path, 'wb') as f:
        f.write(encrypted_binary)
    t3 = time()
    get_file_size(path, eta = t3 - t2)
    t_print('done')
    return None

def decrypte_and_pickle_load(path: str, key: str):
    t_print('f.read()')
    t0 = time()
    with open(path, 'rb') as f:
        data = f.read()
    t1 = time()
    get_file_size(path, eta = t1 - t0)
    t_print('xor_decrypt()')
    decrypted_binary = xor_decrypt(data, key=key, output_format='binary')
    t_print('pickle.loads()')
    t2 = time()
    data = pickle.loads(decrypted_binary)
    t3 = time()
    get_file_size(decrypted_binary, eta = t3 - t2)
    t_print('done')
    return data


#%%
# json

# pickle_load = lambda path: pickle.load(open(path, 'rb'))
# pickle_dump = lambda var, path: pickle.dump(var, open(path, 'wb'))



json_dump = lambda d, path: json.dump(d, open(path, 'w', encoding='utf-8', newline='\n'), indent=4, ensure_ascii=False)
json_dumps = lambda d, indent=4: json.dumps(d, indent=indent, ensure_ascii=False)
json_load = lambda path: json.load(open(path, 'r', encoding='utf-8'))
json_loads = lambda s: json.loads(s)


def touch(path:str):
    if not os.path.exists(path):
        with open(path, 'w'):
            pass
    return None


def read_big_pickle(path: str, interval = [2, 5]):
    if os.path.exists('./read_big_pickle'):
        print('busy')
        count_down(interval)
        df = read_big_pickle(path)
    else:
        print('waiting')
        count_down(interval)
        if not os.path.exists('./read_big_pickle'):
            touch('./read_big_pickle')
            print('reading')
            t0 = time()
            df = pd.read_pickle(path)
            t1 = time()
            get_file_size(path, eta=t1 - t0)
            if os.path.exists('./read_big_pickle'):
                os.remove('./read_big_pickle')
            print('done')
        else:
             df = read_big_pickle(path)
    return df


def var_to_pickle(var, path):
    t0 = time()
    pickle_dump(var, path)
    t1 = time()
    get_file_size(file_path=path, verbose=True, eta=t1 - t0)
    return None


def dump_tmp_pickle(var, var_name = 'var'):
    uuid_hex = get_uuid_hex()
    save_dir = './trash/tmp/var/'
    os.makedirs(save_dir, exist_ok=True)
    pickle_name = f'{uuid_hex}-pickle'
    pickle_path = os.path.join(save_dir, pickle_name)
    print(f"\npickle_path = '{pickle_path}'")
    print(f'{var_name} = pd.read_pickle(pickle_path)')
    var_to_pickle(var, path=pickle_path)
    return pickle_path


def dict_to_json(d, digit = None):
    r = {}
    if isinstance(d, str) or isinstance(d, int) or d is None:
        r = d
    elif type(d).__module__ == 'numpy' and type(d).__name__[:3] == 'int':
        r = dict_to_json(int(d), digit=digit)
    elif type(d).__module__ == 'numpy' and type(d).__name__[:5] == 'float':
        r = dict_to_json(float(d), digit=digit)
    elif isinstance(d, float):
        if str(d) == 'inf':
            r = 'inf'
        elif str(d) == '-inf':
            r = '-inf'
        elif str(d) == 'nan':
            r = 'nan'
        else:
            if digit is not None:
                d = round(d, digit)
            r = d
    elif type(d) == dict:
        r = {dict_to_json(k, digit = digit): dict_to_json(v, digit = digit) for k, v in d.items()}
    elif type(d) == list:
        r = [dict_to_json(k, digit = digit) for k in d]
    elif type(d) == tuple:
        r = [dict_to_json(k, digit=digit) for k in d]
    elif isinstance(d, pd.DataFrame):
        r = dict_to_json(d.to_dict(), digit = digit)
    elif isinstance(d, pd.Series):
        r = dict_to_json(d.to_dict(), digit = digit)
    elif isinstance(d, pd.Interval):
        r = dict_to_json(d.__str__(), digit = digit)
    elif type(d).__module__ == 'ray.tune.sample':
        r = d.domain_str
    else:
        r = f'{type(d).__module__}.{type(d).__name__}'
        if r not in [
            'builtins.function',
            'functools.partial',
            ]:
            print(f'warning: unknown type found: {r}')
    return r



# def dict_to_json(d):
#     r = {}
#     if type(d) == dict:
#         for k, v in d.items():
#             if isinstance(v, (str, bool, float, int)):
#                 r[k] = v
#             elif type(v) == list:
#                 r[k] = [dict_to_json(kk) for kk in v]
#             elif type(v) == dict:
#                 r[k] = dict_to_json(v)
#             elif type(v).__module__ == 'ray.tune.sample':
#                 r[k] = v.domain_str
#             else:
#                 r[k] = f'{type(v).__module__}.{type(v).__name__}'
#     elif type(d) == list:
#         r = [dict_to_json(k) for k in d]
#     else:
#         r = d
#     return r


#%%
# pandas 

# def check_index(s):
#     if type(s) == pd.core.frame.DataFrame:
#         s = s.index.to_list()
#     s = list(s)
#     s = pd.to_numeric(s).tolist()
#     s_sorted = sorted(s)
#     if s != s_sorted:
#         print('index: warning: index sorted')
#     if s_sorted[0] != 1:
#         print('index: warning: series[0] != 1')
#     if len(s_sorted) != len(set(s_sorted)):
#         print('index: warning: duplicated')
#     if s_sorted == list(range(s_sorted[0], s_sorted[-1] + 1)):
#         print(f'index: OK (from {s_sorted[0]} to {s_sorted[-1]}, len = {len(s)})')
#         r = True
#     else:
#         print(f'index: min = {s_sorted[0]}, max = {s_sorted[-1]}, len = {len(s)}')
#         r = False
#     return r


# def check_columns(s):
#     r = False
#     if type(s) == list:
#         if len(s) == len(set(s)):
#             print('columns: OK')
#             r = True
#         else:
#             print('columns: warning: duplicated')
#     elif type(s) == pd.core.frame.DataFrame:
#         s = s.columns.to_list()
#         r = check_columns(s)
#     return r


# def check_dataframe(s):
#     r1 = check_index(s)
#     r2 = check_columns(s)
#     r = all([r1, r2])
#     return r


def rename_and_drop_cols(df: pd.DataFrame, rename = {}, drop_dup_and_reindex = False):
    if len(rename) == 0:
        rename = {k: k for k in df}
        t_print('rename = ', rename)
        df = None
    else:
        df = df[list(rename)].copy()
        df.rename(rename, axis = 1, inplace=True)
        if drop_dup_and_reindex:
            n1 = len(df)
            df.drop_duplicates(inplace=True)
            df.index = pd.Index(range(1, len(df)+1), name = 'index')
            n2 = len(df)
            t_print(f'nrows: {n1} - {n1 - n2} = {n2}')
        print(df.dtypes)
    return df











def fix_df_dtype(df: pd.DataFrame, verbose = False):
    for col in df:
        if df[col].dtype.name == 'Float64':
            df[col] = df[col].astype('float64')
            if verbose:
                print(f"df[{col}].dtype.name == Float64 and fixed.")
        elif df[col].dtype.name in ['int64', 'Int32', 'int32', ]:
            df[col] = df[col].astype('Int64')
            if verbose:
                print(f"df[{col}].dtype.name == int64 and fixed.")
    check_dataframe(df)
    return df


# def get_str_dict(df):
#     r = {}
#     cnt = 0
#     elapsed_time = 60
#     t0 = time()
#     col_n = df.shape[1]
#     for col_name, col in df.items():
#         have_str = False
#         col_ls = col.to_list()
#         for k in col_ls:
#             if type(k) not in [int, float]:
#                 have_str = True
#                 break
#         r[col_name] = have_str
#         cnt += 1
#         t1 = time()
#         if t1 - t0 > elapsed_time:
#             pcnt = cnt / col_n
#             eta = (t1 - t0) / cnt * (col_n - cnt)
#             elapsed_time += 60
#             print(f'{cnt}/{len(df)} {str(col_name)}, {pcnt:1.3f}, T = {t1 - t0:1.0f}s, n_str = {sum(r.values())}, ETA = {eta:1.0f}s')
#     return r


# def fix_array_str(s):
#     if not pd.isna(s) and not isinstance(s, np.ndarray) and s in ['', '""', "''", '"', "'"]:
#         r = float('nan')
#     else:
#         r = s
#     return r


# def equal_col2col(s1, s2):
#     if not isinstance(s1, pd.Series):
#         s1 = pd.Series(s1)
#     if not isinstance(s2, pd.Series):
#         s2 = pd.Series(s2)
#     s1 = s1.apply(fix_array_str)
#     s2 = s2.apply(fix_array_str)
#     idx = ~ (s1.isna() | s2.isna())
#     if idx.any() and (s1[idx] == s2[idx]).all():
#         r = True
#     else:
#         r = False
#     return r

# def equal_col2df(s, df, mod = 1):
#     if not isinstance(s, pd.Series):
#         s = pd.Series(s)
#     if not isinstance(df, pd.DataFrame):
#         df = pd.DataFrame(df)
#     r = False
#     for col_name, col_s in df.items():
#         if equal_col2col(s, col_s):
#             r = True
#             break
#     if mod == 1:
#         print(f'[\'{s.name}\'] == [\'{col_name}\']')
#     else:
#         if r == True:
#             r = [s.name, col_name]
#         else:
#             r = []
#     return r


# def equal_df2df(df1, df2):
#     if not isinstance(df1, pd.DataFrame):
#         df1 = pd.DataFrame(df1)
#     if not isinstance(df2, pd.DataFrame):
#         df2 = pd.DataFrame(df2)
#     r = True
#     for _, col_s in df1.items():
#         if not equal_col2df(col_s, df2):
#             r = False
#     return r


# def equal_df2df_v2(df1, df2):
#     idx2 = pd.DataFrame(index=df2.index)
#     df1 = pd.merge(idx2, df1, left_index=True, right_index=True, how='inner')
#     if not isinstance(df1, pd.DataFrame):
#         df1 = pd.DataFrame(df1)
#     if not isinstance(df2, pd.DataFrame):
#         df2 = pd.DataFrame(df2)
#     r = []
#     for _, col_s in df1.items():
#         eqs = equal_col2df(col_s, df2, mod=2)
#         if eqs:
#             r.append(eqs)
#     return r


def get_series_dtype(s: pd.Series):
    if pd.api.types.is_numeric_dtype(s):
        r = 'number'
    else:
        r = s.dtype.name
    return r

def is_equal_series(s1: pd.Series, s2: pd.Series, ignore_nan = False):
    tp1 = get_series_dtype(s1)
    tp2 = get_series_dtype(s2)
    r = False
    if tp1 == tp2:
        if ignore_nan:
            s1 = s1[~s1.isna()]
            s2 = s2[~s2.isna()]
        idx = s1.index.intersection(s2.index)
        if len(idx) > 0 and s1.loc[idx[:100]].equals(s2.loc[idx[:100]]):
            r = s1[idx].equals(s2[idx])
    return r


def df_to_dict(df: pd.DataFrame, ignore_nan = False):
    if ignore_nan:
        d = {k: v[~v.isna()] for k, v in df.items()}
    else:
        d = {k: v for k, v in df.items()}
    return d


def find_equal_cols(a_df: pd.DataFrame, b_df = None, verbose = True, ignore_nan = False):
    t0 = time()
    # a_dict = {k: v for k, v in a_df.items()}
    a_dict = df_to_dict(a_df, ignore_nan)
    if b_df is None:
        is_solo = True
        b_df = a_df
        b_dict = a_dict
    else:
        is_solo = False
        # b_dict = {k: v[~v.isna()] for k, v in b_df.items()}
        b_dict = df_to_dict(b_df, ignore_nan)
    r = []
    checked_ls = []
    for a_col in a_df:
        for b_col in b_df:
            if is_solo:
                if a_col == b_col:
                    continue
                if sorted([a_col, b_col]) in checked_ls:
                    continue
                else:
                    checked_ls.append(sorted([a_col, b_col]))
            if is_equal_series(a_dict[a_col], b_dict[b_col], ignore_nan=ignore_nan):
                r.append([a_col, b_col])
    t1 = time()
    if verbose:
        t_print(f'T = {1000*(t1-t0):.0f} ms, common_cols =', r)
    return r


def stat_series(df_col, mod = 1, total = None, to_mdl = {'coverage': .01, 'gini': .01}):
    if total == None:
        total = len(df_col)
    stat_sr = pd.Series(dtype=object)
    stat_sr['col_name'] = str(df_col.name)
    stat_sr['dtype'] = str(df_col.dtype.name)
    stat_sr['length'] = len(df_col)
    stat_sr['total'] = total
    stat_sr['len_nan'] = df_col.isna().sum()
    stat_sr['len_non_nan'] = stat_sr['length'] - stat_sr['len_nan']
    stat_sr['coverage'] = (stat_sr['length'] - stat_sr['len_nan']) / stat_sr['total']
    stat_sr['sample'] = df_col.loc[~df_col.isna()][:10].to_list()
    if mod == 1:
        stat_sr['nunique'] = df_col.nunique()
        value_counts = df_col.value_counts()
        stat_sr['value_counts'] = value_counts.iloc[:10].to_dict()
        gini = 1 - ((value_counts / value_counts.sum())**2).sum() if len(value_counts) > 0 else float('nan')
        stat_sr['gini'] = gini
        stat_sr['to_mdl'] = False
        if df_col.dtype.name in ['Int64', 'int64', 'Int32', 'int32', 'Float', 'float', 'Float64', 'float64', 'datetime64[ns]']:
            stat_sr['min'] = df_col.min()
            stat_sr['max'] = df_col.max()
            if df_col.dtype.name not in ['datetime64[ns]']:
                stat_sr['mean'] = df_col.mean()
                stat_sr['std'] = df_col.std()
            else:
                stat_sr['mean'] = float('nan')
                stat_sr['std'] = float('nan')
            
            if not pd.isna(stat_sr['min']) and stat_sr['min'] < stat_sr['max'] and df_col.dtype.name not in ['datetime64[ns]']:
                if stat_sr['coverage'] >= to_mdl['coverage'] and stat_sr['gini'] >= to_mdl['gini']:
                    stat_sr['to_mdl'] = True
        else:
            stat_sr['min'] = float('nan')
            stat_sr['max'] = float('nan')
            stat_sr['mean'] = float('nan')
            stat_sr['std'] = float('nan')
    return stat_sr


def stat_dataframe(df: pd.DataFrame, mod = 1, total = None, verbose = False, to_mdl = {'coverage': .01, 'gini': .01}):
    stat_ls = []
    t0 = time()
    delta_t_0 = 0
    cnt = 0
    t_print('to_mdl:', to_mdl)
    for col in df:
        if verbose:
            t_print(col)
        df_col = df[col]
        stat_0 = stat_series(df_col, mod=mod, total = total, to_mdl = to_mdl)
        stat_ls.append(stat_0)
        cnt += 1
        t1 = time()
        delta_t = t1 - t0
        if delta_t > 60 and delta_t - delta_t_0 > 60:
            t_print(f'{cnt}/{len(df.columns)}, eta = {delta_t / cnt * (len(df.columns)-cnt):1.0f}s')
            delta_t_0 = delta_t
    stat_df = pd.DataFrame(stat_ls)
    stat_df.index += 1
    stat_df.index.name = 'index'
    t_print('len(to_mdl) =', f"{stat_df['to_mdl'].sum()}/{len(stat_df)}")
    return stat_df


date_pattern = '^(\d{8}|\d{4}-\d{1,2}-\d{1,2}|\d{4}-\d{1,2}-\d{1,2}[\ ]+\d{1,2}:\d{1,2}:\d{1,2}|\d{4}/\d{1,2}/\d{1,2})$'
is_date_func = lambda s: True if type(s) == str and re.findall(date_pattern, s) else False


def is_date_col(df_col):
    is_date = False
    if df_col.isna().all():
        pass
    elif df_col.dtype.name in ['datetime64[ns]']:
        is_date = True
    elif df_col.dtype.name in ['Int64', 'int64', 'Int32', 'int32',  'float', 'float64']:
        is_date = ((df_col > 19500000) & (df_col < 20500000)).all()
    elif df_col.dtype.name in ['object']:
        is_date = df_col.apply(is_date_func).all()
    return is_date


def find_date_col(df: pd.DataFrame, mode = 0):
    date_cols = []
    s_cnt = pd.Series(dtype=int)
    for col in df:
        df_col = df[col]
        df_col = df_col.loc[~ df_col.isna()]
        if len(df_col) > 100:
            df_col_100 = df_col.sample(100, random_state=0)
        else:
            df_col_100 = df_col
        if is_date_col(df_col_100) and is_date_col(df_col):
            date_cols.append(col)
            s_cnt[col] = len(df_col)
    if mode == 1:
        date_cols = [s_cnt.idxmax()]
    return date_cols


# def get_date_col_in_datetime64(df: pd.DataFrame, ret_only_one = False):
#     from datetime import time
#     date_cols = []
#     for col in df:
#         if df[col].dtype.name == 'datetime64[ns]':
#             if (df[col].dt.time == time(0, 0)).all():
#                 date_cols.append(col)
#     if ret_only_one:
#         if len(date_cols) != 1:
#             raise ValueError(f'len(date_cols) == {len(date_cols)}')
#         date_cols = date_cols[0]
#     return date_cols


def df_to_numeric(df: pd.DataFrame, errors = 'ignore'):
    for col in df:
        if df[col].dtype.name in ['datetime64[ns]']:
            continue
        df[col] = pd.to_numeric(df[col], errors=errors)
        if df[col].dtype.name in ['object']:
            df[col] = df[col].astype(str)
    return df


def df_parse_date(df):
    date_cols = find_date_col(df)
    t_print(f'date_cols = {date_cols}')
    for date_col in date_cols:
        df[date_col] = df[date_col].apply(parse_date)
        df[date_col] = df[date_col].astype('datetime64[ns]')
    return df


is_hash = lambda s: True if isinstance(s, str) and re.findall('^[0-9a-f]{16}$|^[0-9a-f]{32}$|^[0-9a-f]{64}$', s) else False

def is_all_whole_days(datetime_series: pd.Series):
    datetime_series = datetime_series.dropna()
    return ((datetime_series.dt.hour == 0) & (datetime_series.dt.minute == 0) & (datetime_series.dt.second == 0) & (datetime_series.dt.microsecond == 0)).all()


def get_hash_cols(df: pd.DataFrame, with_date = None):
    hash_cols = []
    obj_cols = df.select_dtypes(include=['object']).columns
    for col in obj_cols:
        df_col = df[col]
        df_col = df_col.loc[~df_col.isna()]
        if len(df_col) > 100:
            df_col = df_col.sample(100)
        if df_col.apply(is_hash).sum() > 90:
            hash_cols.append(col)
    if with_date is not None:
        date_df = df.select_dtypes(include=['datetime64[ns]'])
        date_df = date_df.loc[:, date_df.apply(is_all_whole_days)]
        coverage_sr = 1 - date_df.isna().mean().sort_values()
        date_col = coverage_sr.index[0]
        t_print('date_df coverage:', coverage_sr)
        if len(coverage_sr) > 1:
            print(f'warning: {len(coverage_sr)} date_cols found and set date_col=\'{date_col}\'')
        hash_cols.append(date_col)
    return hash_cols


is_device_type = lambda s: True if isinstance(s, str) and re.findall('oaid|imei|idfa', s) else False

def get_device_type_cols(df: pd.DataFrame):
    device_cols = []
    obj_cols = df.select_dtypes(include=['object']).columns
    for col in obj_cols:
        df_col = df[col]
        df_col = df_col.loc[~df_col.isna()]
        if len(df_col) > 100:
            df_col = df_col.sample(100)
        if df_col.apply(is_device_type).sum() > 90:
            device_cols.append(col)
    return device_cols


def check_hash_df(df:pd.DataFrame, hash_cols = []):
    if hash_cols == []:
        hash_cols = get_hash_cols(df)
    hash_df = df.loc[:, hash_cols]
    t_print('hash_df:')
    t_print(hash_df)
    r = True
    for col in hash_df:
        df_col = hash_df[col]
        df_col = df_col.loc[~df_col.isna()]
        is_hash_s = df_col.apply(is_hash)
        if not is_hash_s.all():
            t_print(f'warning: found wrong hash in [\'{col}\']:')
            t_print(df_col.loc[~is_hash_s])
            r = False
    t_print('missing:', hash_df.isna().sum())
    return r


def get_hash_date_cols(df:pd.DataFrame, date_col = ''):
    if date_col == '':
        date_col = find_date_col(df, mode=1)
    hash_cols = get_hash_cols(df)
    hash_date_cols = hash_cols + date_col
    return hash_date_cols


def is_index(s: pd.Series):
    r = True
    if s.isna().any():
        r = False
    if r and s.dtype.name in ['float64']:
        s2 = s.astype('int64')
        if (s != s2).any():
            r = False
        else:
            s = s2
    if r and s.dtype.name in ['Int64', 'int64', 'Int32', 'int32', ]:
        if not s.is_unique:
            r = False
    else:
        r = False
    return r


def find_index_cols(df: pd.DataFrame):
    index_cols = []
    for col in df:
        if is_index(df[col].iloc[:100]) and is_index(df[col]):
            index_cols.append(col)
    t_print(f'index_cols = {index_cols}')
    return index_cols


def df_reset_index(df: pd.DataFrame):
    index_cols = find_index_cols(df)
    if len(index_cols) == 0:
        r_df = df
        r_df.index = range(1, len(r_df) + 1)
        t_print('warning: index_cols not found and reset index')
    elif len(index_cols) == 1:
        r_df = df.set_index(keys=index_cols[0])
    elif len(index_cols) > 1:
        t_print(f'warning: found multiple index_cols: {index_cols}')
        t_print(df.loc[:, index_cols])
        keys = index_cols[0]
        if (~df.loc[:, index_cols].T.duplicated()).sum() == 1:
            t_print('all index_cols is the same.')  
        else:
            t_print('warning: all index_cols is NOT the same.')
            min_s = df[index_cols].min().sort_values()
            if min_s.iloc[0] == 0 and min_s.iloc[1] == 1:
                keys = min_s[min_s==1].index[0]
            else:
                keys = min_s[min_s==min_s.min()].index[0]
        t_print(f'set index_col = \'{keys}\'')
        r_df = df.set_index(keys=keys)
    return r_df


get_dtype_dict = lambda df: df.dtypes.apply(lambda d: d.name).to_dict()


def is_big_number(s: pd.Series, th = 10**9):
    s = s.loc[~s.isna()]
    r =  len(s) > 0 and (s.abs() > th).all()
    return r


def is_int64(sr: pd.Series):
    try:
        if sr.isna().all():
            r = False
        else:
            sr.astype('Int64')
            r = True
    except:
        r = False
    return r


def skiprows_by_random(n: int, p = .1):
    if n == 0:
        r = False
    else:
        r = random() > p
    return r

def skiprows_by_step(n: int, step = 10):
    if n == 0:
        r = False
    else:
        r = n % step != 1
    return r


def detect_delimiter(filename):
    sep = ','
    with open(filename, 'r') as file:
        first_line = file.readline()
        if '\t' in first_line:
            sep = '\t'
            t_print('detect_delimiter(): info: \\t found.')
    return sep


def select_featuers_by_coverage(data_df: pd.DataFrame, coverage = .01, n = 2000):
    n = min(n, len(data_df))
    t_print('select_featuers_by_coverage()')
    coverage_sr = 1 - data_df.isna().mean()
    coverage_sr.sort_values(ascending=False, inplace=True)
    if n is not None:
        coverage = max(coverage, coverage_sr.iloc[n])
    n1 = len(coverage_sr)
    valid_cols = coverage_sr[coverage_sr >= coverage].index.to_list()
    valid_cols = [col for col in valid_cols if not col.startswith('Unnamed')]
    valid_cols = sorted(valid_cols)
    n2 = len(valid_cols)
    print(f'cols: {n1} - {n1 - n2} = {n2}')
    print(f'{valid_cols[:10]} ... {valid_cols[-10:]}')
    return valid_cols

get_cols_by_coverage = select_featuers_by_coverage

def get_table_header(path: str, output = 'df'):
    if path.endswith('.csv'):
        df_header = pd.read_csv(path, nrows=0)
    elif path.endswith('.xlsx'):
        df_header = pd.read_table(path, nrows=0)
    if output == 'list':
        df_header = df_header.columns.to_list()
    return df_header


def read_csv_by_rows_randomly(path, prob = .5, multiprocessing_read = True):
    t_print('read_csv_by_rows_randomly()')
    path_str = colorize_str(path, [4, 94])
    t_print('path(s) =', path_str)
    if isinstance(path, str) and os.path.isfile(path):
        # df_header = pd.read_csv(path, nrows=0)
        df_header = get_table_header(path)
        total_rows = sum(1 for line in open(path)) - 1
        random_mask = np.random.choice([True, False], size=total_rows, p=[prob, 1-prob])
        df = pd.read_csv(path, skiprows=lambda i: i > 0 and not random_mask[i-1])
        df.columns = df_header.columns
    elif isinstance(path, list):
        paths = path
        dfs = []
        for k, path in enumerate(paths, start = 1):
            t_print(f'{k}/{len(paths)}')
            dfs.append(read_csv_by_rows_randomly(path, prob = prob))
        t_print('pd.concat()')
        df = pd.concat(dfs)
    t_print(df)
    return df


def read_csv_excel(path: str, sheet_name = 0, format = '', dtype = None, encoding=None, skiprows = None, skiprows_func = None, sep=',', header = 0, index_col = None, usecols = None, nrows = None):
    colorful_path = colorize_str(path, [4, 94])
    if skiprows_func is not None:
        if type(skiprows_func) == float:
            seed(0)
            skiprows = functools.partial(skiprows_by_random, p = skiprows)
        elif type(skiprows_func) == int:
            skiprows = functools.partial(skiprows_by_step, step = skiprows)
    dtype_str = str(dtype)
    if len(dtype_str) > 256:
        dtype_str = dtype_str[:128] + ' ... ' + dtype_str[-128:]
    if path.endswith('.xlsx') or path.endswith('.xls') or format in ['xlsx', 'xls']:
        get_sheet_names(path)
        t_print(f'pd.read_excel(\'{colorful_path}\', dtype = {dtype_str})')
        df = pd.read_excel(path, sheet_name=sheet_name, dtype = dtype, skiprows=skiprows, header=header, index_col=index_col, usecols = usecols, nrows=nrows)
    elif path.endswith('.csv') or format in ['csv']:
        t_print(f'pd.read_csv(\'{colorful_path}\', dtype = {dtype_str})')
        if sep == 'auto':
            sep = detect_delimiter(path)
        df = pd.read_csv(path, dtype = dtype, encoding=encoding, skiprows=skiprows, sep=sep, header=header, index_col=index_col, usecols = usecols, nrows=nrows)
    t_print('origin_shape =', df.shape)
    return df


def read_table(path: str, sheet_name = 0, format = '', encoding=None, skiprows = None, skiprows_func=None, sep=',', header = 0, index_col = None, usecols = None, int2float = False, nrows = None):

    origin_df = read_csv_excel(path = path, sheet_name = sheet_name, format = format, encoding=encoding, skiprows = skiprows, skiprows_func=skiprows_func, sep=sep, header=header, index_col = index_col, usecols = usecols, nrows=nrows)
    dtype_dict = get_dtype_dict(origin_df)

    for col in origin_df:
        if dtype_dict[col] == 'object':
            dtype_dict[col] = 'str'
        elif dtype_dict[col] == 'float64':
            if is_int64(origin_df[col]):
                dtype_dict[col] = 'Int64'
            if is_big_number(origin_df[col]):
                dtype_dict[col] = 'str'
    if int2float:
        dtype_dict = {
            k: 'float64' if (v.startswith('int') or v.startswith('Int') ) else v 
            for k, v in dtype_dict.items()}
    del origin_df
    df = read_csv_excel(path = path, sheet_name = sheet_name, format = format, dtype = dtype_dict, encoding=encoding, skiprows = skiprows, skiprows_func=skiprows_func, sep=sep, header=header, index_col = index_col, usecols = usecols, nrows=nrows)

    return df


def auto_reader(
        path: str, 
        sheet_name = 0, 
        format = '', 
        encoding=None, 
        skiprows = None, 
        skiprows_func = None, 
        sep=',', 
        process = ['dropna', 'date', 'index', 'index_name'], 
        header = 0, 
        index_col = None, 
        make_cache = None, 
        use_cache = True, 
        usecols = None, 
        int2float = False, 
        nrows = None
        ):
    if sheet_name != 0 and os.path.splitext(path)[1] in ['.xlsx', '.xls']:
        pickle_path = f'{os.path.splitext(path)[0]}-{sheet_name}{os.path.splitext(path)[1]}.pickle'
    else:
        pickle_path = f'{path}.pickle'
    if os.path.exists(pickle_path) and use_cache and make_cache != 'f':
        pickle_path_str = colorize_str(pickle_path, [4, 94])
        t_print(f"pd.read_pickle('{pickle_path_str}')")
        t0 = time()
        df = pd.read_pickle(pickle_path)
        t1 = time()
        get_file_size(pickle_path, eta=t1 - t0)
        t_print(f'done, shape = {df.shape}')
    else:
        t0 = time()
        df = read_table(path = path, sheet_name = sheet_name, format = format, encoding=encoding, skiprows = skiprows, skiprows_func=skiprows_func, sep=sep, header=header, index_col=index_col, usecols = usecols, int2float=int2float, nrows=nrows)
        if header == None:
            df.columns = [str(k) for k in df.columns]
        if 'dropna' in process:
            len_1 = len(df)
            df.dropna(how = 'all', inplace=True)
            len_2 = len(df)
            t_print('empty_lines =', len_1 - len_2)
        if 'date' in process:
            df = df_parse_date(df)
        if 'index' in process and index_col is None:
            df = df_reset_index(df)
        if 'index_name' in process:
            if df.index.name in [None, 'Unnamed: 0']:
                t_print(f'index.name: \'{str(df.index.name)}\' -> \'index\'')
                df.index.name = 'index'
        t1 = time()
        get_file_size(path, verbose=True, eta = t1-t0)
        t_print(f'done, T = {t1-t0:1.0f}s, shape = {df.shape}')
        if (make_cache and not os.path.exists(pickle_path)) or make_cache == 'f':
            df.to_pickle(pickle_path)
    return df


def read_tables_in_dir(
        tables_path, 
        pattern = '(xlsx|csv)$', 
        sheet_name = 0,
        axis = 0, 
        sort = False, 
        sep = ',', 
        index_col = None, 
        process = ['dropna', 'date', 'index', 'index_name'], 
        use_cache = True, 
        make_cache = False, 
        usecols = None, 
        skiprows = None, 
        nrows = None, 
        int2float = False, 
        header = 0, 
        encoding = 'utf-8', 
        format = '', 
        renew_index = False
        ):
    if type(tables_path) == str:
        if os.path.isfile(tables_path):
            table_paths = [tables_path]
        elif os.path.isdir(tables_path):
            table_paths = ls_files(tables_path, slash=True, pattern=pattern)
        else:
            raise ValueError(f'path = {table_paths} does not exits.')
    elif type(tables_path) == list:
        table_paths = tables_path
        n1 = len(table_paths)
        table_paths = sorted(set(table_paths))
        n2 = len(table_paths)
        if n1 != n2:
            t_print(f'warning: duplicated paths are found and fixed. len_paths = {n1} -> {n2}')
    df_ls = []
    for k, path in enumerate(table_paths):
        t_print(f'{k+1}/{len(table_paths)}')
        df = auto_reader(path, sheet_name = sheet_name, sep = sep, index_col = index_col, process=process, use_cache=use_cache, make_cache=make_cache, usecols=usecols, skiprows=skiprows, nrows=nrows, int2float=int2float, header=header, encoding=encoding, format=format)
        df_ls.append(df)
        del df
    t_print('pd.concat()')
    a_df = pd.concat(df_ls, axis=axis)
    del df_ls
    if sort:
        a_df.sort_index(inplace = True)
        a_df.sort_index(axis = 1, inplace = True)
    
    if renew_index :
        a_df.index = pd.Index(range(1, len(a_df)+1), name='index')
    
    check_dataframe(a_df, error='ignore')

    return a_df


def head_tail_df(df: pd.DataFrame, n = 1000):
    m = n // 2
    ret = pd.concat([
        df.head(m), df.tail(n - m)
        ])
    return ret


def get_corr_df(df:pd.DataFrame):
    df = df.select_dtypes(include=['number'])
    corr = df.corr()
    sr = corr.stack()
    corr_df = pd.DataFrame()
    corr_df['col1'] = [k[0] for k in sr.index]
    corr_df['col2'] = [k[1] for k in sr.index]
    corr_df['corr'] = sr.values
    corr_df = corr_df.loc[corr_df['col1'] != corr_df['col2'], :]
    corr_df.index = range(1, len(corr_df) + 1)
    corr_df.sort_values(by='corr', inplace=True,ascending=False)
    return corr_df


def get_all_combinations(s, least = 1, maximum = None):
    if len(s) > 16:
        raise ValueError(f'get_all_combinations(): len(s) > 16')
    s = sorted(s)
    r = []
    for k in range(least, len(s) + 1):
        r.extend(list(combinations(s, k)))
    r = sorted([sorted(k) for k in r ])
    if maximum is not None:
        r = [k for k in r if len(k) <= maximum]
    return r

# all_combination = get_all_combinations

def get_cols_comb_corr(corr, col_combination):
    corr_comb = corr.loc[col_combination, col_combination]
    corr_min = corr_comb.min().min()
    corr_max = corr_comb.max().max()
    return corr_min, corr_max


def get_cols_comb_corr_df(df):
    df = df.select_dtypes(include=['number'])
    corr = df.corr()
    cols = df.columns.to_list()
    for col in cols:
        corr.loc[col, col] = float('nan')

    col_combinations = get_all_combinations(cols, least=2)
    corr_comb_df = pd.DataFrame(columns=['col_combination', 'n_cols', 'min_corr', 'max_corr'])

    for col_combination in col_combinations:
        corr_min, corr_max = get_cols_comb_corr(corr, col_combination)
        corr_comb_df.loc[len(corr_comb_df)+1] = {
            'col_combination': col_combination, 
            'n_cols': len(col_combination), 
            'min_corr': corr_min, 
            'max_corr': corr_max
            }
    return corr_comb_df


def shrink_df(df: pd.DataFrame, cols = [], mode = 0):
    intersec_cols = sorted(set(cols) & set(df))
    missing_cols = sorted(set(cols) - set(df))
    t_print(f'len(df.columns) = {len(df.columns)}')
    t_print(f'len(valid_cols) = {len(cols)}')
    t_print(f'len(intersec_cols) = {len(intersec_cols)}')
    t_print(f'len(missing_cols) = {len(missing_cols)}')
    df = df.loc[:, intersec_cols]
    if mode == 0:
        if len(missing_cols) > 0:
            raise ValueError(f'missing_cols = {missing_cols}')
    if mode == 1:
        t_print('fill nan cols')
        nan_df = pd.DataFrame(index=df.index, columns=missing_cols, dtype=float)
        df = pd.concat([df, nan_df], axis=1)
    return df


def merge_reindex(a_df: pd.DataFrame, b_df: pd.DataFrame, left_on=None, right_on=None, both_on = None, left_to_right = None, how = 'inner'):
    if left_to_right == 'auto':
        left_to_right = find_similar_cols(a_df=a_df, b_df=b_df)
        if len(left_to_right) == 0:
            raise ValueError(f'left_to_right is empty.')
    if left_to_right is not None and left_on is None and right_on is None:
        left_on = []
        right_on = []
        for a, b in left_to_right:
            left_on.append(a)
            right_on.append(b)
    len_origin_b = len(b_df)
    if both_on is not None and left_on is None and right_on is None:
        left_on = right_on = both_on
    if type(left_on) == str:
        left_on = [left_on]
    if type(right_on) == str:
        right_on = [right_on]
    a_df = a_df.copy()
    t_print(f'b_df.shape = {b_df.shape}')
    n1 = len(b_df)
    dup = b_df.loc[:, right_on].duplicated()
    b_df = b_df.loc[~dup, :]
    n2 = len(b_df)
    if n1 != n2:
        dup_rate = colorize_str(f'dup_rate = {(n1-n2)/n1} (={n1-n2}/{n1})', 93)
        t_print(f'warning: {dup_rate} and have been fixed.')
    else:
        t_print(colorize_str(f'dup_rate = 0', 93))
    a_df['9527-index'] = a_df.index
    a_df.columns = ['9527-' + k for k in a_df.columns]
    left_on = ['9527-' + k for k in left_on]
    m_df = pd.merge(a_df, b_df, left_on=left_on, right_on=right_on, how=how, validate='many_to_one')
    valid_cols = ~ m_df.columns.str.startswith('9527')
    m_df.index = m_df['9527-9527-index']
    m_df.index.name = a_df.index.name
    m_df = m_df.loc[:, valid_cols]
    m_df.sort_index(inplace=True)
    t_print(f'm_df.shape = {m_df.shape}')
    m_df = fix_df_dtype(m_df)
    if not check_dataframe(m_df):
        t_print(m_df)
        raise ValueError('a false df.')
    t_print(colorize_str(f'fake intersec_rate = {len(m_df) / len_origin_b}', 93))
    t_print(colorize_str(f'fake coverage = {len(m_df) / len(a_df)}', 93))
    return m_df


def sample_non_empty(series: pd.Series, n=10):
    non_empty_series = series.dropna()
    if len(non_empty_series) <= n:
        result = non_empty_series
    else:
        result = non_empty_series.sample(n, random_state=0)
    return result


def is_series_contained(series1: pd.Series, series2: pd.Series):
    contained = series1.isin(series2)
    result = contained.mean()
    if series1.dtype.name == 'datetime64[ns]':
        if not is_all_whole_days(series1):
            result = 0
    elif series2.dtype.name == 'datetime64[ns]':
        if not is_all_whole_days(series2):
            result = 0
    return result

def equal_nan_s(s1: pd.Series, s2: pd.Series):
    case1 = s1 == s2
    case2 = s1.isna() & s2.isna()
    eq_s = case1 | case2
    return eq_s


def find_similarity(a_df: pd.DataFrame, b_df: pd.DataFrame, diff_only = False, left_to_right = 'auto',  how = 'outer', n = 10, p = .5, coverage = .9, intersec_index_only = False):
    if intersec_index_only:
        intersec_index = sorted(set(a_df.index) & set(b_df.index))
        a_df = a_df.loc[intersec_index, :]
        b_df = b_df.loc[intersec_index, :]
    if left_to_right == 'auto':
        left_to_right = find_similar_cols(a_df=a_df, b_df=b_df, how=how, n = n, p = p, coverage=coverage)
    common_idx = sorted(set(a_df.index) & set(b_df.index))
    a_len = len(a_df)
    b_len = len(b_df)
    print('len_a =', a_len)
    print('len_b =', b_len)
    a_df = a_df.loc[common_idx, :]
    b_df = b_df.loc[common_idx, :]
    stat_cols = ['same_cnt']
    a_cnt_df = pd.DataFrame(index=[k[0] for k in left_to_right], columns=stat_cols)
    b_cnt_df = pd.DataFrame(index=[k[1] for k in left_to_right], columns=stat_cols)
    for a_col_name, b_col_name in left_to_right:
        a_col = a_df[a_col_name]
        b_col = b_df[b_col_name]
        sim_cnt = equal_nan_s(a_col, b_col).sum()
        a_cnt_df.loc[a_col_name, 'same_cnt'] = sim_cnt
        b_cnt_df.loc[b_col_name, 'same_cnt'] = sim_cnt
    a_cnt_df.insert(0, 'total', a_len)
    b_cnt_df.insert(0, 'total', b_len)
    for cnt_df in [a_cnt_df, b_cnt_df]:
        cnt_df['diff_cnt'] = cnt_df['total'] - cnt_df['same_cnt']
        cnt_df.insert(0, 'col_name', cnt_df.index)
        cnt_df.index = pd.Index(range(1, len(cnt_df)+1), name='index')
        cnt_df['rate'] = cnt_df['same_cnt'] / cnt_df['total']
    if diff_only:
        a_cnt_df = a_cnt_df.loc[a_cnt_df['diff_cnt'] != 0, :]
        b_cnt_df = b_cnt_df.loc[b_cnt_df['diff_cnt'] != 0, :]
    t_print('left_df =', a_cnt_df)
    t_print('right_df =', b_cnt_df)
    return a_cnt_df, b_cnt_df


def find_similar_cols(a_df: pd.DataFrame, b_df: pd.DataFrame, how = 'outer', n = 10, p = .5, coverage = .9):
    a_coverage_sr = 1 - a_df.isna().mean()
    b_coverage_sr = 1 - b_df.isna().mean()
    a_df = a_df.loc[:, a_coverage_sr >= coverage]
    b_df = b_df.loc[:, b_coverage_sr >= coverage]
    a_df = a_df.select_dtypes(exclude='number')
    b_df = b_df.select_dtypes(exclude='number')
    similar_cols = []
    for a_col in a_df:
        for b_col in b_df:
            a_sr = a_df[a_col]
            b_sr = b_df[b_col]
            if a_sr.dtype.name != b_sr.dtype.name:
                continue
            a_sample_sr = sample_non_empty(a_sr, n = n)
            b_sample_sr = sample_non_empty(b_sr, n = n)
            if how == 'left':
                flag = is_series_contained(b_sample_sr, a_sr)
            elif how == 'right':
                flag = is_series_contained(a_sample_sr, b_sr)
            elif how == 'inner':
                flag1 = is_series_contained(b_sample_sr, a_sr)
                flag2 = is_series_contained(a_sample_sr, b_sr)
                flag = min([flag1, flag2])
            elif how == 'outer':
                flag1 = is_series_contained(b_sample_sr, a_sr)
                flag2 = is_series_contained(a_sample_sr, b_sr)
                flag = max([flag1, flag2])
            if flag >= p:
                similar_cols.append([a_col, b_col])
    t_print('similar_cols:', similar_cols)
    return similar_cols


def get_sheet_names(path: str):
    t0 = time()
    t_print(f'load_workbook("{path}", read_only=True)')
    wb = load_workbook(path, read_only=True)
    sheet_names = wb.sheetnames
    t1 = time()
    t_print(f'T = {t1-t0:1.1f}s, sheet_names =', sheet_names)
    return sheet_names


def encrypt_cols(df: pd.DataFrame, prefix = '', suffix = '', verbose = True, n = 8):
    if prefix == suffix == '':
        print('warning: prefix and suffix is empty')
    rename = {k: prefix + get_hash(k, n = n) + suffix for k in df}
    df.rename(rename, axis = 1, inplace = True)
    if verbose:
        t_print('rename =', rename)
    return df, rename

def drop_nan(df: pd.DataFrame, axis = 12, columns = None, coverage = None):
    not_nan_df = ~ df.isna()
    a_idx = not_nan_df.any(axis = 0)
    b_idx = not_nan_df.any(axis = 1)
    if axis == 12:
        input_shape = df.shape
        df = df.loc[b_idx, a_idx]
        output_shape = df.shape
        print(' input:', input_shape)
        print('output:', output_shape)
        print('  diff:', (
            input_shape[0] -  output_shape[0], 
            input_shape[1] -  output_shape[1], 
            ))
    if columns is not None:
        n1 = df.shape[1]
        idx = ~ df[columns].isna().all(axis = 1)
        df = df.loc[idx, :]
        n2 = df.shape[1]
        t_print(f'drop by nan_cols: {n1} - {n1 - n2} = {n2}')
    if coverage is not None:
        coverage_s = (~df.isna()).mean()
        n1 = df.shape[1]
        print(coverage_s)
        high_coverage_cols = coverage_s[coverage_s >= coverage].index
        df = df[high_coverage_cols]
        n2 = df.shape[1]
        t_print(f'drop by coverage: {n1} - {n1 - n2} = {n2}')
        t_print(f'min_coverage = {(~df.isna()).mean().min():.6f}')
    return df


def series_multiprocessing_apply(s, func, json_normalize = False, usecpu = .9):
    t_print('series_multiprocessing_apply()')
    t0 = time()
    n = cpu_count()
    n_in_use = int(usecpu*n)
    t_print(f'cpu_count = {n}, cpu_in_use_count = {n_in_use}')
    pool = Pool(n_in_use)
    s2 = pool.map(func, s)
    pool.close()
    s2 = pd.Series(s2, index=s.index)
    t1 = time()
    t_print(f'T = {t1 - t0:1.3f}s when len = {len(s)}')
    t_print(f'rate = {len(s) / (t1 - t0):.3f} items/s')
    t_print('pd.json_normalize()')
    if json_normalize:
        df = pd.json_normalize(s2)
        df.index = s2.index
        df = df[sorted(df)]
        ret = df
    else:
        ret = s2
    t2 = time()
    t_print(f'T_total = {t2 - t0:.3f}s')
    return ret


def save_num_df(data_df: pd.DataFrame, save_path = None, drop = True, index_range = []):
    data_df = data_df.select_dtypes(include='number')
    if drop:
        data_df = drop_nan(data_df)
        check_dataframe(data_df=data_df, index_range=index_range)
    if save_path is not None:
        t0 = time()
        t_print('data_df.to_pickle(save_path)')
        data_df.to_pickle(save_path)
        t1 = time()
        get_file_size(save_path, eta = t1 - t0)
    return data_df


#%%

import multiprocessing


def multiprocessing_apply(func, vars: object, max_processes = .9, disable_multiprocessing = False, json_normalize = False):
    t_print('multiprocessing_apply()')
    t0 = time()

    if isinstance(vars, pd.DataFrame):
        vars = {col: vars[col] for col in vars.columns}

    if isinstance(vars, pd.Series):
        t_print(f'pd.Series -> dict')
        t0_to_dict = time()
        series_index = vars.index
        vars = vars.to_dict()
        t1_to_dict = time()
        t_print(f'to_dict() T = {t1_to_dict - t0_to_dict:1.3f}s', )
        is_series = True
    else:
        is_series = False

    if not disable_multiprocessing:
        n = cpu_count()
        if max_processes <= 1:
            max_processes = int(max_processes*n)
        else:
            max_processes = int(max_processes)
        t_print(f'cpu_count = {n}, n_concurrent = {max_processes}')
        pool = Pool(max_processes)
    if isinstance(vars, list):
        if disable_multiprocessing:
            t_print(f'warning: disable_multiprocessing = {disable_multiprocessing}')
            results = [func(var) for var in vars]
        else:
            results = pool.map(func, vars)
    elif isinstance(vars, dict):
        keys = list(vars.keys())
        values = [vars[key] for key in keys]
        if disable_multiprocessing:
            t_print(f'warning: disable_multiprocessing = {disable_multiprocessing}')
            value_results =[func(value) for value in values]
        else:
            value_results = pool.map(func, values)
        results = dict(zip(keys, value_results))
    if not disable_multiprocessing:
        pool.close()
    t1 = time()
    t_print(f'T = {t1 - t0:1.3f}s when len = {len(vars)}')
    t_print(f'rate = {len(vars) / (t1 - t0 + 10e-6):.3f} items/s')
    t2 = time()
    if is_series:
        t_print(f'dict -> pd.Series')
        t0_to_series = time()
        results = pd.Series(results)
        t1_to_series = time()
        t_print(f'to_series() T = {t1_to_series - t0_to_series:1.3f}s', )

        if json_normalize:
            t_print('pd.json_normalize()')
            t0_json_normalize = time()
            results = pd.json_normalize(results)
            results.index = series_index
            results = results[sorted(results)]
            t1_json_normalize = time()
            t_print(f'json_normalize() T = {t1_json_normalize - t0_json_normalize:1.3f}s', )
    t_print(f'end of multiprocessing_apply(), T_total = {t2 - t0:.3f}s')
    return results


def stat_dataframe_multiprocessing(df: pd.DataFrame, max_processes = .9, disable_multiprocessing = False):
    cols = [df[col] for col in df]
    stat_df = multiprocessing_apply(
        stat_series, 
        cols, 
        max_processes=max_processes, 
        disable_multiprocessing=disable_multiprocessing
        )
    stat_df = pd.DataFrame(stat_df)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name='index')
    return stat_df


def stat_features_df_by_col(data_df: pd.DataFrame, by_col: str, max_processes = .9, disable_multiprocessing = False):
    t_print('# stat_features_df_by_col()')
    items = sorted(data_df[by_col].unique().tolist())
    num_cols = data_df.select_dtypes(include='number').columns.to_list()
    stat_dict = {}
    for k, item in enumerate(items, start=1):
        t_print(f'## {k}/{len(items)} item =', item)
        idx = data_df[by_col] == item
        part_df = data_df.loc[idx, num_cols]
        stat_dict[item] = stat_dataframe_multiprocessing(part_df, max_processes = max_processes, disable_multiprocessing=disable_multiprocessing)
        print()
    t_print('items =', items)
    t_print('end of stat_features_df_by_col()')
    return stat_dict

def get_qcut_bins_auto(data_sr: pd.Series):
    return get_qcut_bins(data_sr, mode='auto')


def get_features_bins_dict(data_df: pd.DataFrame, disable_multiprocessing = False):
    t_print('get_features_bins_dict()')
    n1 = data_df.shape[1]
    data_df = data_df.select_dtypes(include='number')
    n2 = data_df.shape[1]
    t_print(f'select_dtypes: n_cols: {n1} - {n1 - n2} = {n2}')
    features_bins_dict = multiprocessing_apply(
            get_qcut_bins_auto, 
            {col: data_df[col] for col in data_df},
            disable_multiprocessing=disable_multiprocessing
            )
    return features_bins_dict


def get_col_distri_df(args):
    data_sr, bins_dict = args
    cnt_sr = cal_bins_and_cnts(data_sr, bins=bins_dict[data_sr.name])
    distri_df = pd.DataFrame({'cnt': cnt_sr, 'ratio': cnt_sr / cnt_sr.sum()})
    return distri_df


def get_distri_by_col_dict(data_df: pd.DataFrame, by_col: str, disable_multiprocessing = False):
    t_print('# get_distri_by_col_dict()')
    features_bins_dict = get_features_bins_dict(data_df=data_df, disable_multiprocessing=disable_multiprocessing)
    items = sorted(data_df[by_col].unique())

    distri_by_col_dict = {}
    for item in items:
        print()
        idx = data_df[by_col] == item
        args_list = [(data_df.loc[idx, col], features_bins_dict) for col in features_bins_dict]
        distri_by_col_dict[item] = multiprocessing_apply(
            func=get_col_distri_df, 
            vars={col: args for col, args in zip(features_bins_dict, args_list)},
            disable_multiprocessing=disable_multiprocessing
        )
        del idx
    t_print('# end of get_distri_by_col_dict()')
    return distri_by_col_dict


def cal_psi_value(a_ratio_sr: pd.Series, b_ratio_sr: pd.Series):
    df = pd.DataFrame({
        'a': a_ratio_sr.copy().replace(0, 1e-6),
        'b': b_ratio_sr.copy().replace(0, 1e-6)
    }).fillna(1e-6)
    psi_sr = (df['a'] - df['b']) * np.log(df['a'] / df['b'])
    psi_sum = psi_sr.sum()
    return psi_sum


def get_feature_psi(args):
    distri_by_col_dict, a_item, b_item, feature = args
    psi = cal_psi_value(
        distri_by_col_dict[a_item][feature]['ratio'], 
        distri_by_col_dict[b_item][feature]['ratio']
        )
    return psi


def stat_features_psi_by_col(distri_by_col_dict: dict, max_processes = .9, disable_multiprocessing =False):
    t_print('# stat_features_psi_by_col()')
    items = sorted(set(distri_by_col_dict))
    psi_dict = {}

    total = int(len(items) * (len(items) - 1) / 2)
    t_print('total =', total)
    cnt = 0

    for a_item in items:
        for b_item in items:
            if not(a_item < b_item):
                continue
            cnt += 1
            a_month_dict = distri_by_col_dict[a_item]
            b_month_dict = distri_by_col_dict[b_item]
            features = sorted(set(a_month_dict) & set(b_month_dict))
            if not (len(features) == len(a_month_dict) == len(b_month_dict)):
                raise ValueError
            args_list = [(distri_by_col_dict, a_item, b_item, feature) for feature in features]
            print()
            key = f'psi_btw_{a_item}_and_{b_item}'
            t_print(f'## {cnt}/{total} {key}')
            psi_dict[key] = multiprocessing_apply(
                get_feature_psi, 
                {k: v for k, v in zip(features, args_list)}, 
                max_processes=max_processes,
                disable_multiprocessing=disable_multiprocessing
                )

    psi_df = pd.DataFrame(psi_dict)
    psi_copy_df = psi_df.copy()
    psi_df['min_psi'] = psi_copy_df.min(axis=1)
    psi_df['mean_psi'] = psi_copy_df.mean(axis=1)
    psi_df['max_psi'] = psi_copy_df.max(axis=1)
    psi_df['std_psi'] = psi_copy_df.std(axis=1)
    psi_df['rsd_psi'] = psi_df['std_psi'] / psi_df['mean_psi']

    psi_df = save_and_reset_index(psi_df, 'col_name')
    t_print('end of stat_features_psi_by_col()')
    return psi_df


def get_info_value_dict(data_df: pd.DataFrame, by_col: str, disable_multiprocessing = False):
    t_print('# get_info_value_dict()')
    items = sorted(data_df[by_col].unique())
    info_value_dict = {}
    for k, item in enumerate(items, start=1):
        t_print(f'## {k}/{len(items)} item = {item}')
        item_df = data_df.loc[data_df['month']==item, :]
        info_value_dict[item] = get_distri_by_col_dict(item_df, by_col='y', disable_multiprocessing=disable_multiprocessing)
        print()
    return info_value_dict


def get_feature_info_value(args):
    info_value_dict, item, feature = args
    info_value = cal_psi_value(
        info_value_dict[item]['0'][feature]['ratio'], 
        info_value_dict[item]['1'][feature]['ratio']
        )
    return info_value



def stat_features_info_value_by_col(info_value_dict: dict, max_processes=.9, disable_multiprocessing=True):
    t_print('# stat_features_info_value_by_col()')
    items = sorted(set(info_value_dict))
    iv_dict = {}
    for k, item in enumerate(items, start=1):
        pos_dict = info_value_dict[item]['1']
        neg_dict = info_value_dict[item]['0']
        features = sorted(set(pos_dict) & set(neg_dict))
        if not (len(features) == len(pos_dict) == len(neg_dict)):
            raise ValueError
        args_list = [(info_value_dict, item, feature) for feature in features]
        print()
        key = f'info_value_of_{item}'
        t_print(f'# {k}/{len(items)} {key}')
        iv_dict[key] = multiprocessing_apply(
            get_feature_info_value,
            {k: v for k, v in zip(features, args_list)},
            max_processes=max_processes,
            disable_multiprocessing=disable_multiprocessing
        )
    iv_df = pd.DataFrame(iv_dict)
    iv_copy_df = iv_df.copy()
    iv_df['min_info_value'] = iv_copy_df.min(axis=1)
    iv_df['mean_info_value'] = iv_copy_df.mean(axis=1)
    iv_df['std_info_value'] = iv_copy_df.std(axis=1)
    iv_df['rsd_info_value'] = iv_df['std_info_value'] / iv_df['mean_info_value']
    iv_df['max_info_value'] = iv_copy_df.max(axis=1)
    iv_df = save_and_reset_index(iv_df, origin_index_name='col_name')
    return iv_df


#%%
# model

# def get_sample_xy(n_samples = 10**3, n_features = 5):

#     pos_rate = .4

#     X, y = make_classification(
#         n_samples=n_samples, 
#          n_features=n_features, 
#         #  n_informative=15, n_redundant=5,
#         weights=[1 - pos_rate, pos_rate],
#         random_state=0, )

#     X = pd.DataFrame(X)
#     X.index += 1
#     y = pd.Series(y, index=X.index)
#     X.columns = [f'x_{k}' for k in X.columns]

#     return X, y


# def custom_metric_alpha(X_val, y_val, estimator, labels, X_train, y_train,
#                   weight_val=None, weight_train=None, config=None,
#                   groups_val=None, groups_train=None, alpha = .5):

#     y_train_pred = estimator.predict_proba(X_train)
#     train_loss = log_loss(y_train, y_train_pred, labels=labels, sample_weight=weight_train)
    
#     y_val_pred = estimator.predict_proba(X_val)
#     val_loss = log_loss(y_val, y_val_pred, labels=labels, sample_weight=weight_val)
    
#     # alpha = .5
#     minimized = val_loss + alpha*(val_loss - train_loss) 

#     logged = {
#         "train_loss": train_loss, 
#         "val_loss": val_loss, 
#         'train_auc': roc_auc_score(y_train, y_train_pred[:, 1]),
#         'val_auc': roc_auc_score(y_val, y_val_pred[:, 1]),
#         'train_ks': ks_score(y_train, y_train_pred[:, 1]),
#         'val_ks': ks_score(y_val, y_val_pred[:, 1]),

#         'train_len': len(y_train), 
#         'val_len': len(y_val), 
#         'train_pos_rat': sum(y_train) / len(y_train),
#         'val_pos_rat': sum(y_val) / len(y_val),
#     }
    
#     return minimized, logged


# def get_custom_metric(alpha = .5):
#     return functools.partial(custom_metric_alpha, alpha = alpha) 

# iter_global = 0

# def cv_score_agg_func_alpha(val_loss_folds, log_metrics_folds, alpha = 0):
#     metric_to_minimize = np.mean(val_loss_folds) + alpha*np.std(val_loss_folds)
#     metrics_to_log = log_metrics_folds
#     d = {
#         'train_auc': [], 
#         'val_auc': [], 
#         'train_ks': [], 
#         'val_ks': [], 
#         }
#     for f in log_metrics_folds:
#         for m in d:
#             d[m].append(f[m])
#     rel_err = (np.mean(d['train_ks']-np.mean(d['val_ks']))) / np.mean(d['val_ks']) if np.mean(d['val_ks']) != 0 else float('nan')
#     for k in d:
#         d[k] = f'{np.mean(d[k]):.3f}+-{np.std(d[k]):.3f}' 
#     global iter_global
#     iter_global += 1
#     d['rel_err'] = f'{rel_err:.3f}'
#     d['iter'] = iter_global
#     d['exp_name'] = exp_name_global
#     t_print(d)
#     return metric_to_minimize, metrics_to_log


# def get_cv_score_agg_func(alpha = 0):
#     return functools.partial(cv_score_agg_func_alpha, alpha = alpha) 


# def get_clf_tag(clf):
#     if isinstance(clf, GradientBoostingClassifier):
#         r = 'GB'
#     elif isinstance(clf, HistGradientBoostingClassifier):
#         r = 'HGB'
#     elif isinstance(clf, LGBMClassifier):
#         r = 'LGBM'
#     elif isinstance(clf, XGBClassifier):
#         r = 'XGB'
#     elif isinstance(clf, str):
#         if clf.startswith('GradientBoostingClassifier'):
#             r = 'GB'
#         elif clf.startswith('HistGradientBoostingClassifier'):
#             r = 'HGB'
#         elif clf.startswith('LGBMClassifier'):
#             r = 'LGBM'
#         elif clf.startswith('XGBClassifier'):
#             r = 'XGB'
#         elif clf.startswith('AutoML'):
#             r = 'AutoML'
#         else:
#             raise ValueError(f'clf = {clf}')
#     return r 




def ks_score(y, y_proba, sample_weight = None):
    fpr, tpr = roc_curve(y, y_proba, sample_weight=sample_weight)[:2]
    r = np.abs(tpr - fpr).max()
    return r


def get_auc_ks(y, y_proba, sample_weight = None, dtype = 'dict'):
    r = {}
    t0 = time()
    y, y_proba = to_numpy([y, y_proba])
    r['auc'] = roc_auc_score(y, y_proba, sample_weight=sample_weight)
    r['ks'] = ks_score(y, y_proba, sample_weight=sample_weight)
    if dtype == 'pd.Series':
        r = pd.Series(r)
    t1 = time()
    delta_t = t1 - t0
    if delta_t > 60:
        t_print(f'warning: get_auc_ks() takes too long, T = {delta_t:1.0f}s')
    return r



def get_qcut_bins(series: pd.Series, q = 10, mode = None):
    series = series[~series.isna()]
    if mode == 'auto' and len(series) > 0:
        uniques = sorted(series.unique().tolist())
        if len(uniques) <= 10:
            bins = uniques
        else:
            cnt_sr = series.value_counts()
            cnt_sr = cnt_sr / cnt_sr.sum()
            bins1 = cnt_sr[cnt_sr >= .05].index.to_list()
            bins2 = cnt_sr[cnt_sr < .05].index.to_list()
            new_series = series[series.isin(bins2)]
            if len(new_series) > 0:
                new_bins = pd.qcut(new_series, q=q, duplicates='drop', retbins=True)[1].tolist()
                bins = sorted(set(bins1) | set(new_bins))
            else:
                bins = sorted(bins1)
    elif len(series) > 0:
        bins = pd.qcut(series, q=q, duplicates='drop', retbins=True)[1]
        bins = bins.tolist()
    elif len(series) == 0:
        bins = []
    bins = sorted(bins)
    return bins


def cal_bins_and_cnts(series: pd.Series, bins = 10, percent=False, ret = 'series'):
    if type(bins) == int and bins > 0:
        bins = get_qcut_bins(series, q = bins)
    bins = [-float('inf')] + bins + [float('inf')]
    cut_sr = pd.cut(series, bins = bins)
    cnt_sr = cut_sr.value_counts().sort_index()
    cnt_sr['Not_a_Number'] = series.isna().sum()
    if percent :
        cnt_sr = cnt_sr / cnt_sr.sum()
    r = cnt_sr
    if ret == 'dict':
        cnt_sr.index = cnt_sr.index.to_series().apply(lambda s: s.__str__())
        cnt_dict = cnt_sr.to_dict()
        r = cnt_dict
    return r


def cal_psi(a_cnt_sr, b_cnt_sr):
    a_cnt_sr = a_cnt_sr.copy()
    b_cnt_sr = b_cnt_sr.copy()
    a_pct_sr = a_cnt_sr / a_cnt_sr.sum()
    b_pct_sr = b_cnt_sr / b_cnt_sr.sum()
    for sr in [a_pct_sr, b_pct_sr]:
        for k, v in sr.items():
            if v == 0:
                sr[k] = 1e-6
    cnt_df = pd.DataFrame({
        'a_cnt': a_cnt_sr, 
        'b_cnt': b_cnt_sr, 
        'a_pct': a_pct_sr, 
        'b_pct': b_pct_sr, 
        })
    cnt_df['psi'] = (a_pct_sr - b_pct_sr) * np.log(a_pct_sr / b_pct_sr)
    psi = cnt_df['psi'].sum()
    print(cnt_df)
    print('psi =', psi)
    return cnt_df, psi


def get_psi_by_gp_and_col_mean(grouped, col_sr, flat = True, less_digit = True, bins_mode = None):
    col_name = col_sr.name
    print('\n\n')
    t_print(f'col_name = \'{col_name}\'', )
    bins = get_qcut_bins(col_sr, mode=bins_mode)
    dstr_total = cal_bins_and_cnts(col_sr, bins=bins)
    dstr_by_gp = grouped[col_name].apply(
        lambda s: [
            cal_bins_and_cnts(s, bins=bins), 
            None])
    psi_by_gp = dstr_by_gp.apply(
        lambda dstr_gp: pd.Series(dict(zip(
            ['distr', 'psi'], 
            cal_psi(dstr_total, dstr_gp[0])))))
    if flat:
        new_psi_by_gp = pd.Series(dtype=object)
        for col in psi_by_gp:
            for idx in psi_by_gp.index:
                value = psi_by_gp.loc[idx, col]
                if less_digit and isinstance(value, pd.DataFrame):
                    value = json_dumps(dict_to_json(value.round(3).T), indent=None)
                new_psi_by_gp[f'{col}-{idx}'] = value
        psi_by_gp = new_psi_by_gp
    return psi_by_gp


def get_iv_by_gp_and_col(grouped, col_sr, less_digit = True, bins_mode = None):
    col_name = col_sr.name
    bins = get_qcut_bins(col_sr, mode=bins_mode)
    dstr_by_gp = grouped[col_name].apply(
        lambda s: [
            cal_bins_and_cnts(s, bins=bins), 
            None])
    neg_distr = dstr_by_gp[0][0]
    pos_distr = dstr_by_gp[1][0]
    distr, psi = cal_psi(neg_distr, pos_distr)
    iv_by_gp = pd.Series({
        'neg_distr': neg_distr,
        'pos_distr': pos_distr,
        'distr': distr,
        'info_value': psi})
    for key, value in iv_by_gp.items():
        if less_digit and (isinstance(value, pd.DataFrame) or isinstance(value, pd.Series)):
            value = json_dumps(dict_to_json(value.round(3).T), indent=None)
        iv_by_gp[key] = value
    return iv_by_gp


def get_df_psi_by_gp(data_df, grouped, to_mdl = .1, mode = 'mean', bins_mode = None):
    if mode.lower() in ['mean', 'psi']:
        def func(col_sr):
            return get_psi_by_gp_and_col_mean(grouped, col_sr, bins_mode = bins_mode)
        psi_by_gp_df = data_df.apply(func).T
    elif mode.lower() in ['iv', 'info_value']:
        def func(col_sr):
            return get_iv_by_gp_and_col(grouped, col_sr, bins_mode = bins_mode)
        psi_by_gp_df = data_df.apply(func).T
    psi_by_gp_df.insert(0, 'col_name', psi_by_gp_df.index)
    psi_by_gp_df.index = pd.Index(range(1, len(psi_by_gp_df)+1), name='index')
    
    if mode.lower() in ['mean', 'psi']:
        psi_cols = psi_by_gp_df.columns.str.startswith('psi-')
        psi_by_gp_df['max_psi'] = psi_by_gp_df.loc[:, psi_cols].max(axis = 1)
        psi_by_gp_df['to_mdl'] = psi_by_gp_df['max_psi'] < to_mdl
    elif mode.lower() in ['iv', 'info_value']: 
        psi_by_gp_df['to_mdl'] = psi_by_gp_df['info_value'] >= to_mdl
    t_print('len(to_mdl) =', f"{psi_by_gp_df['to_mdl'].sum()}/{len(psi_by_gp_df)}")
    return psi_by_gp_df


def get_to_mdl_cols(path, mode = None, th = None):
    print(f'mode = {mode}, th = {th}')
    stat_df = pd.read_csv(path, index_col=0, dtype={'col_name': str})
    len_stat_df = len(stat_df)
    if mode is None:
        # print(stat_df)
        cols = stat_df['col_name'][stat_df['to_mdl']].to_list()
        print(f'len(cnt_cols) = {len(cols)}/{len_stat_df}')
    elif mode.lower() in ['iv', 'info_value']:
        iv_bins = pd.qcut(stat_df['info_value'], q=20, duplicates='drop', retbins=True)[1]
        iv_bins = pd.Series(iv_bins).sort_values(ascending=False)
        iv_bins.index = pd.Index(range(len(iv_bins)), name = 'index')
        t_print(iv_bins)
        if th is not None:
            stat_df = stat_df.loc[stat_df['info_value'] >= th, :]
        else:
            stat_df.sort_values(by='info_value', ascending=False, inplace=True)
            stat_df = stat_df.iloc[:1000, :]
        min_iv = stat_df['info_value'].min()
        print('min_info_value =', min_iv)
        cols = stat_df['col_name'].sort_values().to_list()
        print(f'len(iv_cols) = {len(cols)}/{len_stat_df}')
    elif mode.lower() in ['psi']:
        psi_bins = pd.qcut(stat_df['max_psi'], q=20, duplicates='drop', retbins=True)[1]
        psi_bins = pd.Series(psi_bins).sort_values(ascending=True)
        psi_bins.index = pd.Index(range(len(psi_bins)), name = 'index')
        t_print(psi_bins)
        if th is not None:
            stat_df = stat_df.loc[stat_df['max_psi'] < th, :]
        else:
            stat_df = stat_df.loc[stat_df['to_mdl'], :]
        min_max_psi = stat_df['max_psi'].max()
        print('max_max_psi =', min_max_psi)
        cols = stat_df['col_name'].sort_values().to_list()
        print(f'len(psi_cols) = {len(cols)}/{len_stat_df}')
    
    return cols


#%%


# def x_fillna(X: pd.DataFrame, clf_tag, fillna = {}):

#     t0 = time()
#     if fillna == {} :
#         if type(clf_tag) != str:
#             clf_tag = get_clf_tag(clf_tag)
#         if clf_tag == 'LR':
#             fillna = X.mean().to_dict()
#             if X.isna().any().any():
#                 X = X.fillna(value=fillna, axis=0)
#         else:
#             fillna = {k: -10 for k in X}
#             if X.isna().any().any():
#                 X = X.fillna(-10)
#     elif X.isna().any().any():
#         X = X.fillna(value=fillna, axis=0)
#         if X.isna().any().any():
#             X.fillna(-10, inplace=True)
#             fillna2 = {k: fillna[k] if k in fillna else -10 for k in X}
#             fillna = fillna2
#     t1 = time()
#     dt = t1 - t0
#     if dt > 60:
#         t_print(f'x_fillna() took too long for delta_t = {dt:1.0f}s')

#     return X, fillna


# def get_clf(clf_tag: str, params: dict):

#     params2 = params.copy()

#     if 'metric_alpha' in params2:
#         alpha = params2['metric_alpha']
#         del params2['metric_alpha']
#         params2['metric'] = get_custom_metric(alpha)

#     if 'cv_score_agg_func_alpha' in params2:
#         alpha = params2['cv_score_agg_func_alpha']
#         del params2['cv_score_agg_func_alpha']
#         params2['cv_score_agg_func'] = get_cv_score_agg_func(alpha)

#     if clf_tag in ['XGB']:
#         params2['verbosity'] = 1
#     elif clf_tag in ['AutoML']:
#         params2['verbose'] = 3
#     else:
#         params2['verbose'] = 1

#     if clf_tag == 'GB':
#         clf = GradientBoostingClassifier(**params2)
#     elif clf_tag == 'HGB':
#         clf = HistGradientBoostingClassifier(**params2)
#     elif clf_tag == 'LGBM':
#         clf = LGBMClassifier(**params2)
#     elif clf_tag == 'XGB':
#         clf = XGBClassifier(**params2)
#     elif clf_tag == 'LR':
#         clf = LogisticRegression(**params2)
#     elif clf_tag == 'AutoML':
#         clf = AutoML(**params2)
#     else:
#         t_print('No clf_tag found.')

#     if 'random_state' in dir(clf) and clf.random_state == None:
#         clf.random_state = 0
#     elif 'random_state' not in dir(clf):
#         t_print(f'warning: no random_state in {clf}')

#     return clf


# def modeling(X_train: pd.DataFrame, y_train: pd.Series, params = {}, clf_tag = 'GB', fillna = {}, weight = None, XY3 = None):

#     # X_train, fillna = x_fillna(X_train, clf_tag=clf_tag, fillna=fillna)
#     t_print('to_numpy()')
#     t_print(f'X_train\n{X_train}')
#     X_train = X_train.to_numpy()
#     t_print(f'y_train\n{y_train}')
#     y_train = y_train.to_numpy()
#     if weight is not None:
#         if isinstance(weight, pd.Series):
#             weight = weight.to_numpy()

#     t0 = time()
#     if 'hyperopt' not in clf_tag:
#         clf = get_clf(clf_tag=clf_tag, params=params)
#         t_print('clf.fit(X_train, y_train)')
#         clf.fit(X_train, y_train, weight)

#     else:
#         X3, Y3 = XY3
#         t_print('clf = hyperopt_func_v0(X3=X3, Y3=Y3, **params)')
#         clf = hyperopt_func_v0(X3=X3, Y3=Y3, **params)

#     t1 = time()

#     fit_time = t1 - t0
#     if fit_time < 1:
#         sleep(1 - fit_time)

#     t_print(f'fit_time = {fit_time:.1f}s')
#     clf_str = clf.__str__().replace('\n', '').replace(' ', '').replace(',', ', ')
#     t_print(clf_str)

#     clf_json = {}
#     clf_json['clf_str'] = clf_str
#     clf_json['clf_tag'] = clf_tag
#     clf_json['clf_params'] = dict_to_json(params)
#     clf_json['fit_time'] = fit_time
#     clf_json['get_params'] = dict_to_json(clf.get_params(deep=False))

#     return clf, clf_json, fillna


def get_stat_metric(clf, X, y, sample_weight = None):
    if sample_weight is not None and len(sample_weight) == 0:
        sample_weight = None
    data_json = {}
    data_json['stat'] = {}
    data_json['stat']['x'] = {}
    data_json['stat']['y'] = {}
    data_json['stat']['x']['shape'] = list(X.shape)
    data_json['stat']['y']['shape'] = list(y.shape)
    if len(X) == len(y) > 0:
        y_proba = clf.predict_proba(X.to_numpy())[:, 1]
        data_json['stat']['y']['count'] = y.value_counts().to_dict()
        data_json['stat']['y']['percent'] = (y.value_counts()/len(y)).to_dict()
        data_json['metric'] = get_auc_ks(y, y_proba, sample_weight=sample_weight)
    else:
        data_json['stat']['y']['count'] = {'0': 0, "1": 0}
        data_json['stat']['y']['percent'] = {'0': float('nan'), "1": float('nan')}
        data_json['metric'] = {'auc': float('nan'), "ks": float('nan')}
    return data_json


def get_mdl_type_sr(ya_df, y_col, date_col, test_date, stratify, validate_size = .2):
    ya_df = ya_df.loc[~ya_df[y_col].isna(), :]
    mdl_type_sr = pd.Series(index=ya_df.index, dtype=str)
    if type(test_date) == str:
        test_date = datetime.datetime.strptime(test_date, '%Y-%m-%d')
    mdl_type_sr.loc[ya_df[date_col] >= test_date] = 'test'
    ya_train_validate_df = ya_df.loc[mdl_type_sr.isna(), :]
    train_size = 1 - validate_size
    ya_train_df = sample_rows(ya_train_validate_df, train_size, stratify=stratify)
    mdl_type_sr.loc[ya_train_df.index] = 'train'
    mdl_type_sr.loc[mdl_type_sr.isna()] = 'validate'
    mdl_type_sr.name = 'mdl_type'
    return mdl_type_sr


def stat_ya_df(ya_df, date_col, y_col = 'y', mdl_type_col = 'mdl_type'):
    
    y_idx = ~ ya_df[y_col].isna()
    mdl_type_idx = ya_df[mdl_type_col].isin(['train', 'validate', 'test'])
    if not y_idx.equals(mdl_type_idx):
        raise ValueError('mdl_type did not match y_col.')
    
    y_df = ya_df.loc[y_idx, :]
    not_y_df = ya_df.loc[~y_idx, :]

    df = pd.DataFrame(index=['total', 'with_label', 'without_label'])
    df['size'] = {
        'total': len(ya_df),
        'with_label': len(y_df), 
        'without_label': len(not_y_df)
        }
    df['ratio'] = df['size'] / len(ya_df)
    df['min_date'] = {
        'total': ya_df[date_col].min(), 
        'with_label': y_df[date_col].min(), 
        'without_label': not_y_df[date_col].min()
        }
    df['max_date'] = {
        'total': ya_df[date_col].max(), 
        'with_label': y_df[date_col].max(), 
        'without_label': not_y_df[date_col].max()
        }
    print('# with and without label:')
    print(df)
    print()
    
    g1 = y_df.groupby(by=[y_col])
    g11 = g1.agg({date_col: ['min', 'max']})
    g11.columns = ['min_date', 'max_date']
    df = pd.DataFrame({'size': g1.size(), 'ratio': g1.size()/len(y_df)})
    df = pd.concat([df, g11], axis = 1)
    print('# 1-col ratio:', )
    print(df.loc[[0, 1], :])
    print('sum:', )
    print(df[['size', 'ratio']].sum())

    print()
    g2 = y_df.groupby(by=[mdl_type_col])
    g21 = g2.agg({date_col: ['min', 'max']})
    g21.columns = ['min_date', 'max_date']
    df = pd.DataFrame({'size': g2.size(), 'ratio': g2.size()/len(y_df)})
    df = pd.concat([df, g21], axis = 1)
    df = df.loc[['train', 'validate', 'test'], :]
    print(df)
    print('sum:\n')
    print(df[['size', 'ratio']].sum())

    print('\n# 2-col ratio:')
    g3 = y_df.groupby(by=[mdl_type_col, y_col, ])
    g4 = g3.size() / g2.size()
    df = pd.DataFrame({'size': g3.size(), 'ratio': g4})
    df = df.loc[['train', 'validate', 'test'], :]
    print(df)
    print('sum:\n', df.sum())
    return None


def stat_y_by_time(ya_df: pd.DataFrame, by_table = [], time_col = 'month', y_col = 'y', prefix = None):
    if len(by_table) == 0:
        gp2 = ya_df.groupby(by=by_table + [time_col])
        df = pd.DataFrame({
            'size': gp2[y_col].size(),
            'count': gp2[y_col].count(),
            'coverage': gp2[y_col].count() / gp2[y_col].size(),
            'pos_cnt': gp2[y_col].sum(),
            'pos_ratio': gp2[y_col].mean(),
            'pct_by_group': gp2[y_col].size() / len(ya_df)
        })
    else:
        gp1 = ya_df.groupby(by=by_table)
        gp2 = ya_df.groupby(by=by_table + [time_col])
        df = pd.DataFrame({
            'size': gp2[y_col].size(),
            'count': gp2[y_col].count(),
            'coverage': gp2[y_col].count() / gp2[y_col].size(),
            'pos_cnt': gp2[y_col].sum(),
            'pos_ratio': gp2[y_col].mean(),
            'pct_by_group': gp2[y_col].size() / gp1[y_col].size()
        })
    df.sort_index(inplace=True)
    if prefix is not None:
        df = df.add_prefix(prefix)
    return df


def stat_y_by_data_split(ya_df: pd.DataFrame, by_table = [], data_split_col = 'data_split', y_col = 'y', date_col = 'date'):
    ya_df = ya_df.copy()
    if len(by_table) == 0:
        ya_df.insert(0, 'table', 'no_table')
        by_table = ['table']
    ya_df[data_split_col].fillna('NaN', inplace=True)
    gp1 = ya_df.groupby(by=by_table)
    gp2 = ya_df.groupby(by=by_table + [data_split_col])
    df = pd.DataFrame({
        'size': gp2[y_col].size(),
        'count': gp2[y_col].count(),
        'coverage': gp2[y_col].count() / gp2[y_col].size(),
        'pos_cnt': gp2[y_col].sum(),
        'pos_ratio': gp2[y_col].mean(),
        'pct_by_split': gp2[y_col].size() / gp1[y_col].size(),
        'min_date': gp2[date_col].min(),
        'max_date': gp2[date_col].max(),
    })
    df.sort_index(inplace=True)
    return df


# def simple_split_idx(df: pd.DataFrame, y_col = 'y', test_size = .2):
#     print(f'test_size = {test_size}')
#     y_train, y_validate = train_test_split(
#         df, df[y_col],
#         test_size=test_size, 
#         stratify=df[y_col], 
#         random_state=0
#         )[2:]
#     train_idx = y_train.sort_index().index.to_list()
#     validate_idx = y_validate.sort_index().index.to_list()
#     return train_idx, validate_idx


# def append_data_split_col(data_df: pd.DataFrame, test_date: str, y_col = 'y', test_size = .2, date_col = 'date', data_split_col = 'data_split'):
#     if isinstance(test_date, str):
#         test_date = datetime.datetime.strptime(test_date, '%Y-%m-%d') 
#     data_df = data_df.loc[~data_df[y_col].isna(), :].copy()
#     data_df[data_split_col] = pd.Series(dtype=str)
#     test_idx = data_df[date_col] >= test_date
#     data_df.loc[test_idx, data_split_col] = 'test'
#     train_validate_df = data_df.loc[data_df[data_split_col].isna(), :]
#     if test_size == 'auto':
#         test_size = test_idx.sum() / (~test_idx).sum()
#         t_print(f'test_size = {test_size}')
#     train_idx, validate_idx = simple_split_idx(train_validate_df, y_col=y_col, test_size=test_size)
#     data_df.loc[train_idx, data_split_col] = 'train'
#     data_df.loc[validate_idx, data_split_col] = 'validate'

#     gp1 = data_df.groupby(by=[data_split_col])
#     gp2 = data_df.groupby(by=[data_split_col, y_col])
#     stat_df = pd.DataFrame({
#         'size': gp1[y_col].size(),
#         'count': gp1[y_col].count(),
#         'pos_ratio': gp1[y_col].mean(),
#         'min_date': gp1[date_col].min(),
#         'max_date': gp1[date_col].max(),
#         'pct_by_split': gp1[y_col].count() / len(data_df),
#     })
#     print(stat_df)
#     stat_df = pd.DataFrame({
#         'size': gp2[y_col].size(),
#         'count': gp2[y_col].count(),
#         'rate': gp2[y_col].count() / gp1[y_col].size(),
#         'pct_by_split': gp2[y_col].size() / gp1[y_col].size(),
#         'min_date': gp2[date_col].min(),
#         'max_date': gp2[date_col].max(),
#     })
#     print(stat_df)
#     return data_df


# def get_data_split_col(data_df: pd.DataFrame, test_date: str, y_col = 'y', test_size = .2, date_col = 'date', data_split_col = 'data_split'):
#     data_split_col_sr = append_data_split_col(data_df, test_date, y_col, test_size, date_col, data_split_col = data_split_col)[data_split_col]
#     return data_split_col_sr


# def get_xy_tri(X, y, ds = 0, test_size = .2):
#     if len(X) == len(y) == 2:
#         X_sample, X_test = X
#         y_sample, y_test = y
#         X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     elif len(X) == len(y) == 3:
#         X_train, X_valid, X_test = X
#         y_train, y_valid, y_test = y
#         if ds != 0:
#             del test_size
#             test_size = len(y_valid) / (len(y_train) + len(y_valid))
#             t_print('(estimated) test_size =', test_size, ', random_stat =', ds)
#             X_sample = pd.concat([X_train, X_valid]).sort_index()
#             y_sample = pd.concat([y_train, y_valid]).sort_index()
#             X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     elif len(X) == len(y) > 3:
#         X_sample, X_test, y_sample, y_test = train_test_split(X, y, test_size=test_size, random_state=0, stratify=y)
#         X_train, X_valid, y_train, y_valid = train_test_split(X_sample, y_sample, test_size=test_size, random_state=ds, stratify=y_sample)
#     return X_train, X_valid, X_test, y_train, y_valid, y_test


def split_xy_data(y_df: pd.DataFrame, x_df: pd.DataFrame, how = 'intersec', fillna = None, found_nan='raise'):
    
    t_print('fillna =', fillna)
    y_df = y_df.loc[~y_df['data_split'].isna(), :]
    idx = y_df.index.intersection(x_df.index)
    n1 = len(y_df)
    n2 = len(idx)
    t_print(f'coverage: {n2 / n1} (= {n2}/{n1})')
    y_df = y_df.loc[idx, :]
    y_df['y'] = y_df['y'].astype(int)
    x_df = x_df.loc[idx, :].astype(float)
    
    if isinstance(fillna, str) and fillna == 'mean':
        fillna = x_df.loc[y_df['data_split'] == 'train', :].mean()
        t_print('fillna =', fillna.to_dict())
    if fillna is not None:
        x_df = x_df.fillna(fillna)
    
    split_ls = ['train', 'validate', 'test']
    y_dfs = [y_df.loc[y_df['data_split'] == split, 'y'] for split in split_ls]
    x_dfs = [x_df.loc[y_df['data_split'] == split, :  ] for split in split_ls]
    
    t_print(f'y_dfs = {y_dfs}', )
    t_print(f'x_dfs = {x_dfs}', )
    
    check_xy_data(x_dfs, y_dfs, found_nan=found_nan)
    
    return y_dfs, x_dfs


# def modeling_tri(X, y, settings = {}, fillna = {}, params = {}, mode = None, weight = None, more_func = lambda *s: {}):

#     clf_tag = settings['clf_tag']

#     ds = settings['DS.random_state']
#     test_size = settings['DS.test_size']
#     X_train, X_valid, X_test, y_train, y_valid, y_test = get_xy_tri(X, y, ds, test_size = test_size)
#     del X
#     del y

#     if 'selected_dim' in settings:
#         t_print('warning: selected_dim in setting')
#         X_train = X_train.loc[:, settings['selected_dim']]
#         X_valid = X_valid.loc[:, settings['selected_dim']]
#         X_test = X_test.loc[:, settings['selected_dim']]

#     if weight is not None:
#         train_weight, validate_weight, test_weight = weight
#     else:
#         train_weight = validate_weight = test_weight = None

#     if X_train.columns.to_list() == X_valid.columns.to_list() == X_test.columns.to_list():
#         keys = X_train.columns.to_list()

#     X3 = [X_train, X_valid, X_test]
#     Y3 = [y_train, y_valid, y_test]

#     clf, clf_json, fillna = modeling(X_train, y_train, params=params, clf_tag=clf_tag, fillna=fillna, weight=train_weight, XY3=[X3, Y3])

#     all_json = {
#         'main': settings,
#         'clf': clf_json,
#         'data_set': {
#             'train': get_stat_metric(clf, X_train, y_train, sample_weight = train_weight),
#             'validate': get_stat_metric(clf, X_valid, y_valid, sample_weight = validate_weight),
#             'test': get_stat_metric(clf, X_test, y_test, sample_weight = test_weight),
#         },
#         'more': more_func(clf, X3, Y3)
#     }

#     t_print(f"clf_params:\n{json_dumps(all_json['clf']['clf_params'])}")
#     t_print(f"train:\n{json_dumps(all_json['data_set']['train']['metric'])}")
#     t_print(f"validate:\n{json_dumps(all_json['data_set']['validate']['metric'])}")
#     t_print(f"test:\n{json_dumps(all_json['data_set']['test']['metric'])}")

#     return clf, all_json, fillna, keys


short_param = {
    # 'random_state': 'rs',

    # GradientBoostingClassifier
    'learning_rate': 'lr', 
    'n_estimators': 'ne', 
    'subsample': 'ss', 
    'min_samples_split': 'ms', 
    'min_samples_leaf': 'ml', 
    'max_depth': 'md', 
    'max_leaf_nodes': 'mn', 
    
    # HistGradientBoostingClassifier
    'learning_rate': 'lr', 
    'max_iter': 'mi', 
    'max_leaf_nodes': 'mn', 
    'max_depth': 'md', 
    'min_samples_leaf': 'ml', 
    'l2_regularization': 'lt', 

    # LGBMClassifier
    'num_leaves': 'nl', 
    'max_depth': 'md', 
    'learning_rate': 'lr', 
    'n_estimators': 'ne', 
    'min_split_gain': 'mg', 
    'min_child_weight': 'mw', 
    'min_child_samples': 'ml', 
    'subsample': 'ss', 
    'subsample_freq': 'sf', 
    'colsample_bytree': 'ct', 
    'reg_alpha': 'ra', 
    'reg_lambda': 'rl', 
    'class_weight': 'cw',
    'n_jobs': 'nj', 

    # XGBClassifier
    'n_estimators': 'ne', 
    'max_depth': 'md', 
    'max_leaves': 'nl', 
    'learning_rate': 'lr', 
    'n_jobs': 'nj', 
    'min_child_weight': 'mw', 
    'subsample': 'ss', 
    'colsample_bytree': 'ct', 
    'colsample_bylevel': 'cl', 
    'colsample_bynode': 'cn', 
    'reg_alpha': 'ra', 
    'reg_lambda': 'rl', 
    
    # LR
    'penalty': 'pn',
    'class_weight': 'cw',
    'max_iter': 'mi',
    'l1_ratio': 'lo',
    'solver': 'sv',
    'n_jobs': 'nj', 
    'tol': 'tl', 

    # AutoML (possible?!)
    }


LR_solver_dict = {
    'lbfgs': '0',
    'liblinear': '1',
    'newton-cg': '2',
    'newton-cholesky': '3',
    'sag': '4',
    'saga': '5'
    }


def short_num(a):
    if type(a) == str:
        if a == 'balanced':
            r = '1'
        elif a == 'elasticnet':
            r = '12'
        elif a == 'l1':
            r = '1'
        elif a == 'l2':
            r = '2'
        elif a in LR_solver_dict:
            r = LR_solver_dict[a]
    elif type(a) == int or abs(a - round(a)) < 1e-6:
        a = round(a)
        r = str(a)
    elif type(a) == float and a < 1:
        r = str(a).replace('.', '').rstrip('0')
    else:
        print('warning: short_num: stupid type?!', a)
    return r


def params_dict_to_params_str(params):
    d = params.copy()

    for del_key in ['random_state', 'n_jobs']:
        if del_key in d:
            del d[del_key]

    if len(d) > 0:
        s = ''.join([
            short_param[k] + short_num(v)
            for k, v in sorted(d.items())
            if v != None
        ])
    else:
        s = 'default'

    return s


def get_com_dim_name(dims):
    dims = sorted(set(dims))
    if len(dims) == 1:
        dims_suffix = ['-'.join(dims[0].split('-')[1:])]
    else:
        for n in range(max(map(len, dims))):
            if len(set([k[:n] for k in dims])) > 1:
                break
        dims_suffix = [k[n-1:] for k in dims]
    r = 'dim=' + '+'.join(dims_suffix)
    return r


# def fix_settings(settings, X):

#     origin_setting_sha256 = get_sha256(settings, 8)

#     if type(X) == list:
#         X_index_ls = [Xk.index.to_list() for Xk in X]
#         X = X[0]
#     else:
#         X_index_ls = X.index.to_list()
    
#     cols = X.columns.to_list()
#     if 'dim_name_func' in settings:
#         dim_name_func = settings['dim_name_func']
#         # dim_str = get_com_dim_name(settings['selected_dim'])
#         # cols_sha256 = get_sha256(settings['selected_dim'], 8)
#         # dim_str = f'dim_{cols_sha256}'
#         if 'selected_dim' in settings:
#             selected_dim = settings['selected_dim']
#         else:
#             selected_dim = cols
#         dim_str = dim_name_func(selected_dim)
#     elif 'selected_dim' in settings:
#         cols_sha256 = get_sha256(settings['selected_dim'], 8)
#         dim_str = f'dim_{cols_sha256}'
#     else:
#         # idxs_sha256 = get_sha256(X_index_ls, 4)
#         # cols_sha256 = get_sha256(cols, 4)
#         # dim_str = f'dim_{idxs_sha256}_{cols_sha256}'
#         cols_sha256 = get_sha256(cols, 8)
#         dim_str = f'dim_{cols_sha256}'

#     if 'clf_tag' not in settings:
#         settings['clf_tag'] = 'GB'

#     clf_tag = settings['clf_tag']

#     # if clf_tag == 'AutoML':
#     #     t_print('warning: AutoML log will not be saved.')
#     #     if 'exp_name' not in settings:
#     #         exp_name = 'AutoML_' + get_sha256(settings, 8)
#     #         settings['exp_name'] = exp_name
#     #         t_print(f'warning: exp_name was missing and let expname = {exp_name}.')

#     params_keys = [k for k in settings if k.startswith(clf_tag)]
#     params = {k[len(clf_tag)+1:]: settings[k] for k in params_keys}
#     for k in params_keys:
#         del settings[k]

#     if 'DS.random_state' not in settings:
#         settings['DS.random_state'] = 0
#     ds_str = '-ds' + str(settings['DS.random_state'])

#     if 'DS.test_size' not in settings:
#         settings['DS.test_size'] = .2
        
#     ds_str = '-ds' + str(settings['DS.random_state'])


#     if 'random_state' in params:
#         rs = params['random_state']
#     else:
#         rs = 0

#     if 'exp_name' not in settings:
#         prefix = settings['prefix']
#         clf_tag = settings['clf_tag']
#         if clf_tag in ['AutoML', 'hyperopt']:
#             params_str = get_hash(params, n = 8)
#         else:
#             params_str = params_dict_to_params_str(params)
#         exp_name = f'{prefix}-{clf_tag}_{params_str}-{dim_str}'
#         settings['exp_name'] = exp_name
#     elif not settings['exp_name'].startswith(settings['prefix']):
#         settings['exp_name'] = f"{settings['prefix']}-{settings['exp_name']}"
#     exp_name = settings['exp_name']

#     save_name_0 = f'{exp_name}_rs{rs}{ds_str}-'
#     settings['save_name_0'] = save_name_0

#     return settings, params


# def save_result_files(clf, all_json, fillna, keys, add_pmml = True):

#     all_json = dict_to_json(all_json)

#     try:
#         if isinstance(clf, AutoML):
#             clf = clf.model.estimator
#     except:
#         t_print('warning: No AutoML')
#         pass

#     results_path = all_json['main']['results_path']
#     prefix = all_json['main']['prefix']
#     exp_name = all_json['main']['exp_name']
#     save_name_0 = all_json['main']['save_name_0']

#     exp_path = os.path.join(results_path, prefix, exp_name)

#     save_results = True

#     if os.path.exists(exp_path):
#         exps = [k[:-17] for k in os.listdir(exp_path)]
#         if save_name_0 in exps:
#             save_results = False
#             save_name_0_path = os.path.join(exp_path, save_name_0)
#             save_name_0_path = replace_slash(save_name_0_path)
#             t_print(f'{save_name_0_path}* exists')

#     if save_results:
#         timestamp = get_timestamp()
#         save_name = save_name_0 + timestamp
#         save_path = os.path.join(exp_path, save_name)
#         save_path = replace_slash(save_path)
#         all_json['main']['timestamp'] = timestamp
#         all_json['main']['uuid_hex'] = get_uuid_hex()
#         all_json['main']['save_name'] = save_name
#         all_json['main']['save_path'] = save_path
#         os.makedirs(save_path)

#         clf_path = os.path.join(save_path, 'clf_pickle')
#         pickle_dump(clf, clf_path)

#         clf_json_path = os.path.join(save_path, 'clf.json')
#         json_dump(all_json, clf_json_path)

#         fillna = {k: v for k, v in fillna.items() if k in keys}
#         fillna_json_path = os.path.join(save_path, 'fillna.json')
#         json_dump(fillna, fillna_json_path)

#         keys_json_path = os.path.join(save_path, 'keys.json')
#         json_dump(keys, keys_json_path)

#         if add_pmml :
#             add_mdl_files(save_path, timestamp=timestamp)

#         dump_files_hash(save_path, timestamp = timestamp)

#         t_print(f'{save_path} saved')

#     return None


def run_clf(settings):

    r = True

    results_path = settings['results_path']
    prefix = settings['prefix']
    exp_name = settings['exp_name']
    save_name_0 = settings['save_name_0']

    exp_path = os.path.join(results_path, prefix, exp_name)

    if os.path.exists(exp_path):
        exps = [k[:-17] for k in os.listdir(exp_path)]
        if save_name_0 in exps:
            r = False
            save_name_0_path = os.path.join(exp_path, save_name_0)
            save_name_0_path = replace_slash(save_name_0_path)
            t_print(f'{save_name_0_path}* exists')

    return r


def get_params_dict_list(all_params_dict):

    all_keys = list(all_params_dict.keys())
    for k in all_keys:
        if all_params_dict[k] == []:
            del all_params_dict[k]

    keys = all_params_dict.keys()
    values = [k if type(k) == list else [k] for k in all_params_dict.values()]
    params_dict_list = [dict(zip(keys, combination)) for combination in product(*values)]
    
    r = []
    for k in params_dict_list:
        if k not in r:
            if k['clf_tag'] == 'AutoML':
                if k['AutoML.metric_alpha'] * k['AutoML.cv_score_agg_func_alpha'] != 0:
                    continue
            r.append(k)

    return r


# def more_func(clf, X3: list, Y3: list):
#     if 'model' in dir(clf):
#         clf = clf.model.estimator
#     Y3_proba = [
#         pd.Series(clf.predict_proba(xs)[:, 1], index=ys.index)
#         for xs, ys in zip(X3, Y3)
#         ]
#     ret = {}
#     ret['evaluation'] = {}
#     ret['evaluation']['gain_and_lift'] = get_tri_lift(Y3, Y3_proba)
#     ret['evaluation']['distribution_by_interval'] = get_tri_distribution_by_interval(Y3, Y3_proba)
#     ret['evaluation']['cdf_curve'] = get_tri_cdf_curve(Y3, Y3_proba)
#     ret['evaluation']['describe'] = get_tri_describe(Y3, Y3_proba)
#     ret['feature'] = {}
#     ret['feature']['feature_importances'] = get_feature_info(clf, X3, Y3)
#     return ret


def normalize_df(df: pd.DataFrame, verbose = False):
    norm_df = (df - df.mean()) / df.std()
    if verbose:
        t_print('describe =', norm_df.describe().T)
    return norm_df


def get_LR_coef(clf, X_df, normalized = False, to_series = True):
    coef = dict(zip(X_df.columns.to_list(), clf.coef_[0]))
    coef = pd.Series(coef)
    if normalized:
        coef = coef.abs()
        coef = coef / coef.sum()
    return coef


# def get_LR_FI(clf, X_train, y_train):
#     X_norm_train = normalize_df(X_train)
#     norm_clf = LogisticRegression(max_iter = clf.max_iter, random_state = 0)
#     norm_clf.fit(X_norm_train, y_train)
#     describe_df = X_train.describe().T
#     describe_df['count'] = describe_df['count'].astype(int)
#     FI_df = pd.DataFrame({
#         'origin_coef': get_LR_coef(clf, X_train),
#         'norm_coef': get_LR_coef(norm_clf, X_norm_train),
#         'feature_importances': get_LR_coef(norm_clf, X_train, normalized = True),
#     })
#     stat_df = pd.concat([describe_df, FI_df], axis = 1)
#     save_and_reset_index(stat_df, 'feature')
#     return stat_df


# def get_feature_info(clf, X3: list, Y3: list):
#     features = X3[0].columns.to_list()
#     if isinstance(clf, LGBMClassifier):
#         FI_df = get_FI_df(features=features, clf=clf)
#         s_idx = FI_df['split_origin'] != 0
#         g_idx = FI_df['gain_origin'] != 0
#         if s_idx.equals(g_idx):
#             idx = s_idx
#             features_in_use = FI_df['feature'][idx].to_list()
#             features_in_use_cnt = len(features_in_use)
#         d = {'LGBMClassifier': {
#             'feature_importances_stat': FI_df,
#             'features_in_use': features_in_use,
#             'features_in_use_cnt': features_in_use_cnt,}}
#     elif isinstance(clf, XGBClassifier):
#         FI_df = get_FI_df(features=features, clf=clf)
#         idx = FI_df['gain'] != 0
#         features_in_use = FI_df['feature'][idx].to_list()
#         features_in_use_cnt = len(features_in_use)
#         d = {'XGBClassifier': {
#             'feature_importances_stat': FI_df,
#             'features_in_use': features_in_use,
#             'features_in_use_cnt': features_in_use_cnt,}
#             }
#     elif isinstance(clf, LogisticRegression):
#         X_train = X3[0]
#         y_train = Y3[0]
#         FI_stat_df = get_LR_FI(clf, X_train, y_train)
#         d = {'LogisticRegression': {
#             'feature_importances_stat': FI_stat_df,
#         }}
#         t_print(FI_stat_df)
#     return d


def simple_describe(s: np.array):
    s = pd.Series(s)
    d = s.describe().to_dict()
    d['count'] = int(d['count'])
    d['size'] = len(s)
    return d


def get_tri_describe(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': simple_describe(y_proba_train),
        'y_proba_validate': simple_describe(y_proba_valid),
        'y_proba_test': simple_describe(y_proba_test),
        
        'y_proba_train_1': simple_describe(y_proba_train[y_train==1]),
        'y_proba_validate_1': simple_describe(y_proba_valid[y_valid==1]),
        'y_proba_test_1': simple_describe(y_proba_test[y_test==1]),
        
        'y_proba_train_0': simple_describe(y_proba_train[y_train==0]),
        'y_proba_validate_0': simple_describe(y_proba_valid[y_valid==0]),
        'y_proba_test_0': simple_describe(y_proba_test[y_test==0]),
    }
    return d


def get_tri_cdf_curve(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': get_cdf_curve(y_proba_train),
        'y_proba_validate': get_cdf_curve(y_proba_valid),
        'y_proba_test': get_cdf_curve(y_proba_test),
        
        'y_proba_train_1': get_cdf_curve(y_proba_train[y_train==1]),
        'y_proba_validate_1': get_cdf_curve(y_proba_valid[y_valid==1]),
        'y_proba_test_1': get_cdf_curve(y_proba_test[y_test==1]),
        
        'y_proba_train_0': get_cdf_curve(y_proba_train[y_train==0]),
        'y_proba_validate_0': get_cdf_curve(y_proba_valid[y_valid==0]),
        'y_proba_test_0': get_cdf_curve(y_proba_test[y_test==0]),
    }
    return d


def get_tri_distribution_by_interval(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    d = {
        'y_proba_train': get_distribution_by_interval(y_proba_train),
        'y_proba_validate': get_distribution_by_interval(y_proba_valid),
        'y_proba_test': get_distribution_by_interval(y_proba_test),
        
        'y_proba_train_1': get_distribution_by_interval(y_proba_train[y_train==1]),
        'y_proba_validate_1': get_distribution_by_interval(y_proba_valid[y_valid==1]),
        'y_proba_test_1': get_distribution_by_interval(y_proba_test[y_test==1]),
        
        'y_proba_train_0': get_distribution_by_interval(y_proba_train[y_train==0]),
        'y_proba_validate_0': get_distribution_by_interval(y_proba_valid[y_valid==0]),
        'y_proba_test_0': get_distribution_by_interval(y_proba_test[y_test==0]),
    }
    return d


def get_tri_lift(Y3, Y3_proba):
    y_train, y_valid, y_test = Y3
    y_proba_train, y_proba_valid, y_proba_test = Y3_proba
    train_df = get_gain_and_lift(y_train, y_proba_train)
    valid_df = get_gain_and_lift(y_valid, y_proba_valid)
    test_df = get_gain_and_lift(y_test, y_proba_test)
    tri_lift_dict = {
        'train': train_df,
        'validate': valid_df,
        'test': test_df
    }
    return tri_lift_dict

def check_xy_data(X = [], y = [], found_nan = 'raise'):
    for Xk in X:
        print('Xmin =', Xk.min().min())
        print('Xmax =', Xk.max().max())
        if Xk.isna().any().any() and found_nan == 'raise':
            raise ValueError('X has nan')
        elif Xk.isna().any().any():
            print('X has nan')
        else:
            print('X has no nan')
    for yk in y:
        print(yk.value_counts().to_dict())
        if yk.isna().any() :
            raise ValueError('y has nan')
    return None


# exp_name_global = ''


# def auto_params_v2(X, y, settings_dict, weight=None, more_func = more_func, test_only = False, add_pmml = True, fillna = None):

#     if type(X) == tuple:
#         X = list(X)
    
#     if type(y) == tuple:
#         y = list(y)

#     X = [Xk.copy().astype(float) for Xk in X]
#     y = [yk.copy() for yk in y]
    
#     if not (len(X) == len(X) == 3):
#         raise ValueError('X and y must be list of length 3')

#     if isinstance(fillna, int) or isinstance(fillna, float) :
#         fillna = {k: fillna for k in X[0].keys()}

#     if isinstance(fillna, str) and fillna == 'mean':
#         fillna = X[0].mean()

#     if isinstance(fillna, pd.Series):
#         fillna = fillna.to_dict()
#         print('fillna =', fillna)

#     t_print('fillna')
#     [Xk.fillna(fillna, inplace = True) for Xk in X]

#     if not isinstance(fillna, dict):
#         raise ValueError('fillna must be dict or pd.Series')

#     settings_dict_list = get_params_dict_list(settings_dict)
#     shuffle(settings_dict_list)

#     global exp_name_global
#     global iter_global

#     eta = 0
#     t0 = time()
#     for k, settings in enumerate(settings_dict_list):
#         t1 = time()
#         cnt = k + 1
#         t_print(f'{cnt}/{len(settings_dict_list)}')
#         settings, params = fix_settings(settings, X)
#         exp_name_global = settings['exp_name']
        
#         if run_clf(settings):
#             t_print(settings)
#             t_print(params)
#             t_print(f'T_one = 0 at {date_str()}')
#             if not test_only:
#                 sleep(1)
#                 clf, all_json, _, keys = modeling_tri(X, y, settings=settings, params=params, weight=weight, more_func=more_func)
#                 save_result_files(clf, all_json, fillna, keys, add_pmml=add_pmml)
#             else:
#                 t_print('test_only = True')
#             t2 = time()
#             eta = (t2 - t0) / (k+1) * (len(settings_dict_list) - cnt)
#             t_print(f'T_one = {t2 - t1:1.0f}s')
#             iter_global = 0
#             exp_name_global = ''
#         t_print(f'All will end at {date_str(eta)}.\n#')
        
#     stat_df = parse_clfs(os.path.join(settings_dict['results_path'], settings['prefix']))
#     t_print(f'auto_params_v2() ends at {date_str()}.')
#     t_print(f'end of', settings['prefix'])
#     print('\n'*5)
#     return stat_df



#%%
# hyperopt

def to_numpy(var, fillna = None) -> np.array:
    if isinstance(var, pd.DataFrame) or isinstance(var, pd.Series) :
        # if fillna is not None:
        #     var = var.fillna(-10)
        var = var.to_numpy()
    elif isinstance(var, list):
        var = [to_numpy(k) for k in var]
    return var


def loss_func_alpha(metric_dict: dict, loss_alpha = 1, loss_items = None, reversed_as_inf = True):

    reverse = any([value['reverse'] for value in metric_dict.values()])

    validate_auc_loss = 1 - metric_dict['validate']['auc']
    test_auc_loss = 1 - metric_dict['test']['auc']

    validate_perf = metric_dict['train']['auc']-.5 + metric_dict['train']['ks']/2
    test_perf = metric_dict['test']['auc']-.5 + metric_dict['test']['ks']/2
    validate_loss = 1 - validate_perf
    test_loss = 1 - test_perf
    
    auc_overfit = abs(metric_dict['train']['auc'] - metric_dict['validate']['auc'])
    ks_overfit = abs(metric_dict['train']['ks'] - metric_dict['validate']['ks'])
    overfit = (2*auc_overfit + ks_overfit) / 2
    
    auc_decline = abs(metric_dict['validate']['auc'] - metric_dict['test']['auc'])
    ks_decline = abs(metric_dict['validate']['ks'] - metric_dict['test']['ks'])
    decline = (2*auc_decline + ks_decline) / 2

    loss_dict = {
        'validate_auc_loss': validate_auc_loss,
        'test_auc_loss': test_auc_loss,

        'validate_loss': validate_loss, 
        'test_loss': test_loss,
        'overfit': loss_alpha * overfit,
        'decline': loss_alpha * decline,
    }

    if loss_items is None:
        loss = validate_perf + loss_alpha*overfit
    elif loss_items == 'validate_auc':
        loss = (1 - metric_dict['validate']['auc']) + loss_alpha*abs(metric_dict['train']['auc'] - metric_dict['validate']['auc'])
    else:
        t_print('loss_items =', loss_items)
        loss = sum([loss_dict[k] for k in loss_items])

    if reverse and reversed_as_inf:
        loss = float('inf')

    t_print('loss =', loss)

    return loss


def get_clf_auc_ks(clf, X, y, weight=None):
    y_pred = clf.predict_proba(X)[:, 1]
    auc = roc_auc_score(y, y_pred, sample_weight=weight)
    ks = ks_score(y, y_pred, sample_weight=weight)
    ret = {'auc': auc, 'ks': ks, 'reverse': auc <= .5}
    return ret


# def objective_func_with_params(params, XY3 = None, loss_func = None, fillna = None):

#     if XY3 is not None: 
#         XY3 = to_numpy(XY3, fillna=fillna)
#         X3, Y3 = XY3
#         X_train, X_validate, X_test = X3
#         y_train, y_validate, y_test = Y3

#     if True:
#         params['scale_pos_weight'] = (y_train==0).sum() / (y_train==1).sum()

#     t_print('model = XGBClassifier(**', params, ', random_state=0)')
#     model = XGBClassifier(**params, random_state=0)
#     t_print('model.fit(X_train, y_train)')
#     model.fit(X_train, y_train)

#     metric_dict = {
#         'train': get_clf_auc_ks(clf=model, X=X_train, y=y_train),
#         'validate': get_clf_auc_ks(clf=model, X=X_validate, y=y_validate),
#         'test': get_clf_auc_ks(clf=model, X=X_test, y=y_test),
#     }

#     t_print(metric_dict)

#     loss = loss_func(metric_dict)

#     ret = {
#         'loss': loss, 
#         'status': STATUS_OK, 
#         'metric_dict': metric_dict,
#         }
    
#     return ret


def early_stop(trials, eps = .001, n_tail_trials = 10, n_least_trails = 20, *args):
    flag = False
    n_trials = len(trials.trials)
    print(f'n_trials = {n_trials} completed.\n')
    n = 10 # 10
    if n_trials > n_tail_trials:
        losses = [trial['result']['loss'] for trial in trials.trials]
        losses = pd.Series(losses)
        losses_head = losses.iloc[:-n]
        losses_tail = losses.iloc[-n:]
        # losses_range = losses.max() - losses.min()
        t_print('losses_tail.describe() =', losses_tail.describe())
        # if losses.std() < m_losses_stat:
        decline_gap = losses_head.min() - losses_tail.min()
        t_print('decline_gap =', decline_gap)
        if n_trials > n_least_trails and decline_gap < eps:
            t_print(f'early_stop(): flag = True')
            flag = True
    return flag, args


# space = {
#     'n_estimators': hp.choice('n_estimators', [10, 20, 50, 100]),
#     'learning_rate': hp.choice('learning_rate', [0.01, 0.05, 0.1]),
#     'max_depth': hp.choice('max_depth', [1, 2, 3]),
#     # 'subsample': hp.choice('subsample', [0.5, 0.7, 1.0]),
#     'subsample': hp.choice('subsample', [.2, .5, .8, 1.0]),
#     # 'gamma': hp.choice('gamma', [0.0, 0.1, 0.2]),
#     'gamma': hp.choice('gamma', [0, .1, .2, .5, .8, 1.0]),
#     # 'colsample_bytree': hp.choice('colsample_bytree', [0.5, 0.7, 1.0])
#     'colsample_bytree': hp.choice('colsample_bytree', [.2, .5, .8, 1.0])
# }


# space = {
#     'n_estimators': hp.choice('n_estimators', list(range(10, 201, 10))),
#     'learning_rate': hp.uniform('learning_rate', .01, .20),
#     'max_depth': hp.choice('max_depth', [1, 2, 3]),
#     'subsample': hp.uniform('subsample', .1, 1.0),
#     'gamma': hp.uniform('gamma', 0, 5),
#     'colsample_bytree': hp.uniform('colsample_bytree', .1, 1.0),
# }


# try:
#     hyperopt_func_v0_space = {
#         # "n_estimators": hp.choice("n_estimators", [20, 50, 100, 150, 200]),
#         # "n_estimators": hp.randint("n_estimators", 10, 101),
#         "n_estimators": hp.randint("n_estimators", 20, 101),
#         'learning_rate': hp.uniform('learning_rate', .01, .20),
#         # "max_depth": hp.choice("max_depth", [2, 3, 4, 5]),
#         "max_depth": hp.choice("max_depth", [1, 2, 3]),
#         "subsample": hp.uniform("subsample", 0.1, 1),
#         "colsample_bytree": hp.uniform("colsample_bytree", 0.1, 1),
#         "gamma": hp.uniform("gamma", .1, 10),
#         "reg_alpha": hp.uniform("reg_alpha", .1, 10),
#         # "min_child_weight": hp.choice('min_child_weight', [20, 50, 100, 200, 500]),
#         "min_child_weight": hp.randint("min_child_weight", 10, 1001),
#         }
# except:
#     print('warning: hyperopt_func_v0_space is not defined.')
#     pass


# def hyperopt_func_v0(X3: list, Y3: list, space = None, fillna = None, **kwargs):
#     max_evals = kwargs.get('max_evals', 100)
#     loss_alpha = kwargs.get('loss_alpha', 1)
#     loss_items = kwargs.get('loss_items', ['validate_loss', 'overfit'])
#     eps = kwargs.get('eps', .001)
#     n_tail_trials = kwargs.get('n_tail_trials', 10)
#     n_least_trails = kwargs.get('n_least_trails', 20)
#     reversed_as_inf = kwargs.get('reversed_as_inf', True)
#     if space is None:
#         space = hyperopt_func_v0_space
#     objective_func = functools.partial(
#         objective_func_with_params, 
#         XY3 = [X3, Y3],
#         loss_func=functools.partial(
#             loss_func_alpha, 
#             loss_alpha = loss_alpha, 
#             loss_items=loss_items,
#             reversed_as_inf = reversed_as_inf
#             ),
#         fillna = fillna
#         )
#     trials = Trials()
#     best_params = fmin(fn=objective_func,
#                 space=space,
#                 algo=tpe.suggest,
#                 max_evals=max_evals,
#                 trials=trials,
#                 early_stop_fn=functools.partial(
#                     early_stop, 
#                     eps = eps, 
#                     n_least_trails = n_least_trails, 
#                     n_tail_trials = n_tail_trials, 
#                     ),
#                 rstate=np.random.default_rng(0),
#                 )
#     best_params = space_eval(space, best_params)
#     if True:
#         best_params['scale_pos_weight'] = sum(Y3[0] == 0) / sum(Y3[0] == 1)
#     t_print(f're-train model at params =', best_params)
#     best_model = XGBClassifier(**best_params, random_state=0)
#     best_model.fit(X3[0], Y3[0])
#     return best_model


#%%
# model report 

def save_and_reset_index(df: pd.DataFrame, origin_index_name = None):
    if origin_index_name is not None:
        df.index.name = origin_index_name
    df.insert(0, df.index.name, df.index)
    df.index = pd.Index(range(1, len(df)+1), name='index')
    return df


def get_describe_by_month_df(xy_df: pd.DataFrame):
    gp = xy_df.groupby(by=['month'])['y_proba']
    df = gp.describe()
    df.insert(0, 'size', gp.size())
    df.loc['total'] = xy_df['y_proba'].describe()
    df.loc['total', 'size'] = len(xy_df)
    df.insert(2, 'coverage', df['count'] / df['size'])
    df['size'] = df['size'].astype(int)
    df['count'] = df['count'].astype(int)
    df = save_and_reset_index(df)
    return df


def roc_auc_score_with_nan(y: pd.Series, y_score: pd.Series):
    idx = (~y.isna()) & (~y_score.isna())
    if idx.sum() > 1 and len(y[idx].unique()) > 1 > 0:
        ret = roc_auc_score(y[idx], y_score[idx])
    else:
        ret = float('nan')
    return ret

def ks_score_with_nan(y: pd.Series, y_score: pd.Series):
    idx = (~y.isna()) & (~y_score.isna())
    if idx.sum() > 1 and len(y[idx].unique()) > 1 > 0:
        ret = ks_score(y[idx], y_score[idx])
    else:
        ret = float('nan')
    return ret


def get_auc_and_ks_by_data_split(xy_df: pd.DataFrame, rename = None):
    idx = (~xy_df['data_split'].isna())
    data_df = xy_df.loc[idx, :].copy()
    gp = data_df.groupby(by=['data_split'])
    df = pd.DataFrame({
        'min_date': gp['date'].min(),
        'max_date': gp['date'].max(),
        'size': gp['y_proba'].size(),
        'count': gp['y_proba'].count(),
        'pct': float('nan'),
        'pos_cnt': gp['y'].sum(),
        'pos_ratio': gp['y'].mean(),
        'auc': gp.apply(lambda s: roc_auc_score_with_nan(s['y'], s['y_proba'])),
        'ks': gp.apply(lambda s: ks_score_with_nan(s['y'], s['y_proba']))
    })
    df['pct'] = df['size'] / df['size'].sum()
    data_split_idx = ['train', 'validate', 'test']
    df = df.loc[data_split_idx, :]
    df = save_and_reset_index(df)
    if rename is not None:
        df['data_split'] = df['data_split'].map(rename)
    return df


def get_auc_and_ks_by_month(xy_df: pd.DataFrame):
    idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna())
    data_df = xy_df.loc[idx, :].copy()
    gp = data_df.groupby(by=['month'])
    df = pd.DataFrame({
        'min_date': gp['date'].min(),
        'max_date': gp['date'].max(),
        'size': gp['y_proba'].size(),
        'count': gp['y_proba'].count(),
        'coverage': float('nan'),
        'pct': float('nan'),
        'pos_cnt': gp['y'].sum(),
        'pos_ratio': gp['y'].mean(),
        'auc': gp.apply(lambda s: roc_auc_score_with_nan(s['y'], s['y_proba'])),
        'ks': gp.apply(lambda s: ks_score_with_nan(s['y'], s['y_proba']))
    })
    df['coverage'] = df['count'] / df['size']
    df['pct'] = df['size'] / df['size'].sum()
    df = save_and_reset_index(df)
    return df


def get_cut_points(s: pd.Series, bins = 20, endpoints = [-float('inf'), float('inf')]):
    points = []
    s = pd.Series(s).copy()
    s.fillna(s.mean(), inplace=True)
    s.sort_values(inplace=True)
    for k in range(1, bins):
        n = int(len(s) * k / bins)
        cut = s.iloc[n]
        points.append(cut)
    n1 = len(points)
    points = sorted(set(points + endpoints))
    n2 = len(points)
    print(f'get_cut_points: info: bins = {n2-1}')
    if n1 + len(endpoints) != n2:
        print('get_cut_points: warning: too many duplicated numbers.')
    points = sorted(points)
    return points


def get_gain_and_lift(y, y_proba, bins = 'auto', prefix = '', y_proba_max = 1):
    if bins == 'auto':
        bins = get_cut_points(y_proba, bins=20, endpoints=[0, y_proba_max])
    bins = sorted(bins, reverse=True)
    bins = [k for k in bins if k < y_proba_max]
    print('bins =', bins)
    df = pd.DataFrame(
        index=range(1, len(bins)), 
        columns=[
            'interval_left', 'interval_right', 
            'total_count', 'pos_count', 'total_pct', 'pos_ratio', 
            'gain', 'lift',
            ],
        dtype=int)
    y_mean = y.mean()
    y_sum = y.sum()
    for k, cut in enumerate(bins):
        k = k+1
        idx = y_proba > cut
        cut_y = y[idx]
        cnt = len(cut_y)
        cnt_rate = cnt / len(y)
        pos_cnt = cut_y.sum()
        pos_rate = pos_cnt / cnt
        lift = pos_cnt / cnt / y_mean
        # df.loc[k, 'cut_point'] = cut
        df.loc[k, 'interval_left'] = cut
        df.loc[k, 'interval_right'] = y_proba_max
        df.loc[k, 'total_count'] = cnt
        df.loc[k, 'total_pct'] = cnt_rate
        df.loc[k, 'gain'] = pos_cnt / y_sum
        df.loc[k, 'pos_ratio'] = pos_rate
        df.loc[k, 'pos_count'] = pos_cnt
        df.loc[k, 'lift'] = lift
    df = df.loc[df['pos_count'] > 0, :]
    df = df.add_prefix(prefix)
    df.insert(0, 'interval', df.apply(lambda row: f"({row['interval_left']:.9f}, {row['interval_right']:.9f}]", axis=1))
    df.insert(0, 'subset', (((df['total_pct']*100).round()/5).round()*5).apply(lambda s: f'top{s:0.0f}'))
    df['total_count'] = df['total_count'].astype(int)
    df['pos_count'] = df['pos_count'].astype(int)
    df.index.name = 'index'
    print('lift =')
    print(df)
    return df


def get_gain_and_lift_by_data_split(xy_df: pd.DataFrame, rename = None, y_proba_max = 1):
    idx = (~xy_df['y_proba'].isna()) & (~xy_df['y'].isna())
    data_df = xy_df.loc[idx, :]
    dfs = []
    for split in ['train', 'validate', 'test']:
        split_df = data_df.loc[data_df['data_split']==split, :]
        lift_df = get_gain_and_lift(split_df['y'], split_df['y_proba'], y_proba_max = y_proba_max)
        lift_df.insert(0, 'data_split', split)
        dfs.append(lift_df)
    df = pd.concat(dfs)
    df.index = pd.Index(range(1, len(df)+1), name='index')
    if rename is not None:
        df['data_split'] = df['data_split'].map(rename)
    return df


def get_gain_and_lift_by_month(xy_df: pd.DataFrame, y_proba_max = 1):
    idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna()) & (~xy_df['y_proba'].isna())
    data_df = xy_df.loc[idx, :]
    dfs = []
    for month in sorted(data_df['month'].unique()):
        split_df = data_df.loc[data_df['month']==month, :]
        lift_df = get_gain_and_lift(split_df['y'], split_df['y_proba'], y_proba_max = y_proba_max)
        lift_df.insert(0, 'month', month)
        dfs.append(lift_df)
    df = pd.concat(dfs)
    df.index = pd.Index(range(1, len(df)+1), name='index')
    return df


def get_cdf_curve(s: pd.Series):
    if not isinstance(s, pd.Series):
        s = pd.Series(s)
    s = s.copy()
    s.sort_values(inplace=True)
    n = len(s)
    df = pd.DataFrame([{
        'cnt': round(k/99*(n-1)),
        'pct': k/100, 
        'value': s.iloc[round(k/99*(n-1))]
        } for k in range(100)])
    df.index.name = 'index'
    return df


def get_distribution_by_interval(y_proba: pd.Series, bins = 'auto', y_proba_max = 1):
    y_proba = pd.Series(y_proba)
    if bins == 'auto':
        bins = get_cut_points(y_proba, bins=20, endpoints=[0, y_proba_max])
    def func(s):
        sr = pd.cut(s, bins=bins, precision=9).value_counts().sort_index()
        sr['Not_a_Number'] = s.isna().sum()
        return sr
    df = pd.DataFrame({
        'cnt': func(y_proba),
    })
    df['pct'] = df['cnt'] / df['cnt'].sum()
    df.index.name = 'interval'
    df = save_and_reset_index(df)
    df['interval_left'] = df['interval'].apply(lambda s: s.left if isinstance(s, pd.Interval) else float('nan'))
    df['interval_right'] = df['interval'].apply(lambda s: s.right if isinstance(s, pd.Interval) else float('nan'))
    cols = ['interval', 'interval_left', 'interval_right', 'cnt', 'pct']
    df = df[cols]
    return df


def get_distribution_by_data_split(xy_df: pd.DataFrame, rename = None, y_proba_max = 1):
    idx = (~xy_df['data_split'].isna()) #& (~xy_df['y_proba'].isna())
    data_df = xy_df.loc[idx, :].copy()

    bins = get_cut_points(data_df['y_proba'], bins=10, endpoints=[0, y_proba_max])
    gp = data_df.groupby(by=['data_split'])

    def func(s):
        sr = pd.cut(s, bins=bins, precision=9).value_counts().sort_values()
        sr['Not_a_Number'] = s.isna().sum()
        return sr
    
    a_df = pd.DataFrame({
        'count': gp['y_proba'].apply(func),
        })
    a_df['pct'] = a_df['count'] / data_df['data_split'].value_counts()
    a_df.sort_index(inplace=True)

    dfs = []
    for split in ['train', 'validate', 'test']:
        split_df = a_df.loc[split]
        split_df = split_df.add_prefix(f'{split}_')
        dfs.append(split_df)
    b_df = pd.concat(dfs, axis=1)
    b_df.index.name = 'interval'
    b_df = save_and_reset_index(b_df)
    if rename is not None:
        new_rename = {}
        new_rename |= {k+'_count': v+'_count' for k, v in rename.items()}
        new_rename |= {k+'_pct': v+'_pct' for k, v in rename.items()}
        b_df.rename(new_rename, axis=1, inplace=True)
    return b_df


def get_binning_table_by_data_split(xy_df, y_proba_max = 1, rename = None):
    valid_idx = (~xy_df['data_split'].isna()) & (~xy_df['y'].isna()) & (~xy_df['y_proba'].isna())
    xy_df = xy_df.loc[valid_idx, :]
    dfs = []
    for split in ['train', 'validate', 'test']:
        part_df = xy_df.loc[xy_df['data_split']==split, :].copy()
        bins = get_cut_points(part_df['y_proba'], bins=20, endpoints=[0, y_proba_max])
        part_df['interval'] = pd.cut(part_df['y_proba'], bins = bins)
        gp = part_df.groupby(by='interval', observed=True)
        part_stat_df = pd.DataFrame({
            'size': gp.size(),
            'pct': gp.size() / len(part_df),
            'min': gp['y_proba'].min(),
            'max': gp['y_proba'].max(),
            'pos_ratio': gp['y'].mean()
        })
        part_stat_df = save_and_reset_index(part_stat_df, )
        part_stat_df.sort_index(inplace=True)
        part_stat_df.insert(0, 'data_split', split)
        dfs.append(part_stat_df)
    stat_df = pd.concat(dfs)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name = 'index')
    if rename is not None:
        stat_df['data_split'] = stat_df['data_split'].map(rename)
    return stat_df


def get_binning_table_by_month(xy_df, y_proba_max = 1):
    idx = (~xy_df['month'].isna()) & (~xy_df['y'].isna()) & (~xy_df['data_split'].isna()) & (~xy_df['y_proba'].isna())
    xy_df = xy_df.loc[idx, :]
    dfs = []
    months = sorted(xy_df['month'].unique())
    for month in months:
        part_df = xy_df.loc[xy_df['month']==month, :].copy()
        bins = get_cut_points(part_df['y_proba'], bins=20, endpoints=[0, y_proba_max])
        part_df['interval'] = pd.cut(part_df['y_proba'], bins = bins)
        gp = part_df.groupby(by='interval', observed=True)
        part_stat_df = pd.DataFrame({
            'size': gp.size(),
            'pct': gp.size() / len(part_df),
            'min': gp['y_proba'].min(),
            'max': gp['y_proba'].max(),
            'pos_ratio': gp['y'].mean()
        })
        part_stat_df = save_and_reset_index(part_stat_df, )
        part_stat_df.sort_index(inplace=True)
        part_stat_df.insert(0, 'month', month)
        dfs.append(part_stat_df)
    stat_df = pd.concat(dfs)
    stat_df.index = pd.Index(range(1, len(stat_df)+1), name = 'index')
    return stat_df


def get_report_dict(xy_df: pd.DataFrame, rename = None, y_proba_max = 1):
    report_dict = {}
    report_dict['describe_by_month'] = get_describe_by_month_df(xy_df=xy_df)
    report_dict['auc_and_ks_by_data_split'] = get_auc_and_ks_by_data_split(xy_df=xy_df, rename=rename)
    report_dict['auc_and_ks_by_month'] = get_auc_and_ks_by_month(xy_df=xy_df)
    report_dict['gain_and_lift_by_data_split'] = get_gain_and_lift_by_data_split(xy_df=xy_df, rename=rename, y_proba_max = y_proba_max)
    report_dict['gain_and_lift_by_month'] = get_gain_and_lift_by_month(xy_df=xy_df, y_proba_max = y_proba_max)
    report_dict['binning_table_by_data_split'] = get_binning_table_by_data_split(xy_df=xy_df, y_proba_max = y_proba_max, rename = rename)
    report_dict['binning_table_by_month'] = get_binning_table_by_month(xy_df=xy_df, y_proba_max = y_proba_max)
    report_dict['distribution_by_data_split'] = get_distribution_by_data_split(xy_df=xy_df, rename=rename, y_proba_max=y_proba_max)
    return report_dict



#%%
# report

def nan_addition(a, b):
    if pd.isna(a):
        r = b
    elif pd.isna(b):
        r = a
    else:
        r = a + b
    return r


def parse_clf_json(path: str):
    d = json_load(path)
    row = pd.Series(dtype=object)
    row['prefix'] = d['main']['prefix']
    row['exp_name'] = d['main']['exp_name']
    row['save_name'] = d['main']['save_name']
    row['uuid_hex'] = d['main'].get('uuid_hex', get_hash(row['save_name'], alg='md5'))
    row['exp_path'] = os.path.join(
        d['main']['results_path'], 
        row['prefix'],
        row['exp_name']
        ).replace('\\', '/')
    # row['save_path'] = d['main']['save_path']
    row['save_path'] = os.path.dirname(path).replace('\\', '/')
    row['date'] = datetime.datetime.strptime(d['main']['save_name'][-17:], "%Y%m%d%a%H%M%S")
    row['clf_tag'] = d['main']['clf_tag']
    row['clf_str'] = d['clf']['clf_str']
    row['clf_params'] = d['clf']['clf_params']
    row['fit_time'] = round(d['clf']['fit_time'], 1)
    row['len'] = sum([d['data_set'][k]['stat']['x']['shape'][0] for k in ['train', 'validate', 'test']])
    row['dim'] = d['data_set']['train']['stat']['x']['shape'][1]

    row['train_auc'] = d['data_set']['train']['metric']['auc']
    row['validate_auc'] = d['data_set']['validate']['metric']['auc']
    row['test_auc'] = d['data_set']['test']['metric']['auc']
    row['train_ks'] = d['data_set']['train']['metric']['ks']
    row['validate_ks'] = d['data_set']['validate']['metric']['ks']
    row['test_ks'] = d['data_set']['test']['metric']['ks']

    row['abs_overfit_auc'] = row['train_auc'] - row['validate_auc']
    row['rel_overfit_auc'] = row['abs_overfit_auc'] / (row['validate_auc'] - .5) if (row['validate_auc'] - .5) > 0 else float('nan')

    row['abs_decline_auc'] = row['validate_auc'] - row['test_auc']
    row['rel_decline_auc'] = row['abs_decline_auc'] / (row['validate_auc'] - .5) if (row['validate_auc'] - .5) > 0 else float('nan')

    row['abs_auc_overfit_and_decline'] = nan_addition(abs(row['abs_overfit_auc']), abs(row['abs_decline_auc']))

    row['overfit'] = row['rel_overfit_auc']
    row['decline'] = row['rel_decline_auc']

    # row['overfit'] = (row['train_ks'] - row['validate_ks']) / row['validate_ks'] if row['validate_ks'] != 0 else float('nan')
    # row['decline'] = (row['validate_ks'] - row['test_ks'])  / row['validate_ks'] if row['validate_ks'] != 0 else float('nan')
    # row['abs_overfit'] = row['train_ks'] - row['validate_ks']  
    # row['abs_decline'] = row['validate_ks'] - row['test_ks'] 
    # row['abs_overfit_and_decline'] = abs(row['abs_overfit']) + abs(row['abs_decline'])
    
    if 'selected_dim' in d['main']:
        row['selected_dim'] = json.dumps(d['main']['selected_dim'], ensure_ascii=False)
    else:
        row['selected_dim'] = float('nan')

    if 'more' in d:
        row['more'] = compress_and_encode_str(
            json.dumps(
                dict_to_json(d['more'], 6), 
                ensure_ascii=False
                ), 
                tag = True
                )
    else:
        row['more'] = '{}'

    return row


def choose_one_idx(df: pd.DataFrame):
    for n in range(1, len(df)+1):
        idx = set(df.index)
        for col in df:
            m = df[col].sort_values(ascending=False).iloc[:n].min()
            col_idx = df[col].loc[df[col] >= m].index.to_list()
            idx = idx & set(col_idx)
        if len(idx) > 0:
            idx = sorted(idx)
            break
    max_s = []
    for col in df:
        sr = df[col] == df[col].max()
        max_idx = sr.loc[sr].index.to_list()
        max_s.extend(max_idx)
        max_s = sorted(set(max_s))
    idx = sorted(set(idx) | set(max_s))
    return idx


# def neg10(a):
#     if np.isnan(a):
#         r = float('nan')
#     else:
#         if a < .01:
#             a = .01
#         r = - np.floor(1000*a)
#     return r


def round_nan(a):
    if not pd.isna(a):
        r = round(a, 3)
    else:
        r = float('nan')
    return r


def parse_clfs(path: str, save = True):
    paths = ls_dirs_r(path, any_ls='^clf.json$')
    stat_dict = {k: os.path.join(path,'clf.json') for k, path in enumerate(paths, start=1)}
    stat_dict = multiprocessing_apply(parse_clf_json, stat_dict)
    df = pd.DataFrame(stat_dict)
    df = df.transpose()

    df['ks_std'] = df.loc[:, ['train_ks', 'validate_ks', 'test_ks']].std(axis=1)
    df.sort_values(by=['exp_name', 'validate_auc'], ascending=[True, False], inplace=True)

    # df['choose'] = False
    # choose_df = df.loc[:, ['validate_auc', 'test_auc', 'validate_ks', 'test_ks', 'overfit', 'decline']]
    # choose_df['validate_auc'] = (choose_df['validate_auc']).apply(round_nan)
    # choose_df['test_auc'] = (choose_df['test_auc']).apply(round_nan)
    # choose_df['validate_ks'] = (choose_df['validate_ks']).apply(round_nan)
    # choose_df['test_ks'] = (choose_df['test_ks']).apply(round_nan)
    # choose_df['overfit'] = (-choose_df['overfit'].abs()).apply(round_nan)
    # choose_df['decline'] = (-choose_df['decline']).apply(round_nan)
    # choose_idx = choose_one_idx(choose_df)
    # df.loc[choose_idx, 'choose'] = True

    # sort_df = df.loc[:, ['validate_ks', 'test_ks']]#.applymap(lambda a: round(a, 3))
    # sort_df.loc[df['overfit'] >= 1, :] = float('nan')
    # sort_df.sort_values(by=['validate_ks', 'test_ks'], ascending = False, inplace=True)
    # df = df.loc[sort_df.index, :]
    # df.loc[df.index.to_list()[0], 'choose'] = True

    df.index = pd.Index(range(1, len(df) + 1), name='index') 
    if save:
        if type(save) != str:
            save_path = path + '-stat.csv'
        else:
            save_path = save
        df.to_csv(save_path)
        t_print(f'{save_path} saved. len_clfs = {len(df)}')
    show_mdl_stat_info(df)
    return df


def mk_summary(path, pattern = '', unique = False, save = True):
    if type(path) == str:
        path = path.rstrip('/\\')
        paths = ls_dirs(path, pattern=pattern)
    elif type(path) == list:
        paths = path
    df_ls = []
    for k, p in enumerate(paths):
        if os.path.basename(p).startswith('.'):
            t_print(f'skip path = {p}')
            continue
        t_print(f'{k+1}/{len(paths)}')
        a_df = parse_clfs(p)
        dup = a_df.loc[:, ['train_auc', 'validate_auc', 'test_auc', 'train_ks', 'validate_ks', 'test_ks']].duplicated()
        n1 = len(a_df)
        if unique:
            a_df = a_df.loc[~dup, :]
            n2 = len(a_df)
            t_print(f'unique_results: {n1} -> {n2}')
        df_ls.append(a_df)
    df = pd.concat(df_ls)
    df['duplicated'] = df['prefix'].duplicated()
    df.index = pd.Index(range(1, len(df) + 1), name='index') 
    if save :
        df.to_csv(path + '-stat.csv')
    return df


def get_auc_ks_v2(y, y_proba, fillna = 'mean'):
    if fillna == 'mean':
        fillna = y_proba.mean()
    y_proba = y_proba.fillna(fillna).copy()
    d = get_auc_ks(y, y_proba)
    reverse = False
    if d['auc'] < .48:
        y_proba = y_proba.max() - y_proba
        d = get_auc_ks(y, y_proba)
        reverse = True
    elif d['auc'] < .52:
        reverse = float('nan')
    d['reverse'] = reverse
    return d


def get_auc_ks_tri_v2(ya_df: pd.DataFrame, score_sr: pd.Series, fillna = 'mean', y_col = 'y'):
    # score_sr = score_sr.fillna(score_sr.mean()).copy()
    auc_ks_tri = {}
    mdl_types = ['train', 'validate', 'test']
    for mdl_type in mdl_types:
        idx = ya_df['data_split'] == mdl_type
        y = ya_df.loc[idx, y_col]
        y_proba = score_sr[idx]
        d = get_auc_ks_v2(y, y_proba, fillna=fillna)
        auc_ks_tri[f'{mdl_type}_auc'] = d['auc']
        auc_ks_tri[f'{mdl_type}_ks'] = d['ks']
    return auc_ks_tri


def get_coverage_auc_ks_df(ya_df: pd.DataFrame, score_df: pd.DataFrame, y_col = 'y', fillna = 'mean'):
    d_ls = []
    ya_df = ya_df.loc[~ya_df[y_col].isna(), :]
    score_df = score_df.select_dtypes(include='number')
    score_df = pd.merge(
        pd.DataFrame(index=ya_df.index), 
        score_df,
        left_index=True,
        right_index=True,
        how='left',
        validate='one_to_one'
    )
    for score_name in score_df:
        t_print(f'score_name = {score_name}', )
        score_sr = score_df[score_name]
        coverage = 1 - score_sr.isna().sum() / len(score_sr)
        d = {
            'score_name': score_name,
            'len': len(score_sr),
            'coverage': coverage,
        }
        if coverage > 0:
            auc_ks_dict = get_auc_ks_v2(y=ya_df[y_col], y_proba=score_sr, fillna=fillna)
            auc_ks_tri_dict = get_auc_ks_tri_v2(ya_df=ya_df, score_sr=score_sr, fillna=fillna, y_col=y_col)
            d.update(auc_ks_dict)
            d.update(auc_ks_tri_dict)
        d_ls.append(d)
    coverage_auc_ks_df = pd.DataFrame(d_ls)
    coverage_auc_ks_df['overfit'] = coverage_auc_ks_df['train_ks'] / coverage_auc_ks_df['validate_ks'] - 1
    coverage_auc_ks_df['decline'] = 1 - coverage_auc_ks_df['test_ks'] / coverage_auc_ks_df['validate_ks']
    coverage_auc_ks_df.index += 1
    coverage_auc_ks_df.index.name = 'index'
    cols = [
        'score_name',
        'len',
        'coverage',
        'auc',
        'ks',
        'reverse',
        'train_auc',
        'validate_auc',
        'test_auc',
        'train_ks',
        'validate_ks',
        'test_ks',
        'overfit',
        'decline'
        ]
    coverage_auc_ks_df = coverage_auc_ks_df[cols]
    return coverage_auc_ks_df


#%%


def get_clf_path(path = './', verbose = False, n = 0):
    s = os.listdir(path)
    if 'clf.json' in s:
        r = path
        if verbose:
            t_print(f'save_path = {path}')
            json_path = os.path.join(path, 'clf.json')
            clf_json = json_load(json_path)
            if 'main'in clf_json:
                show_json = {k: v for k, v in clf_json['main'].items() if k in ['prefix', 'exp_name']}
                t_print(show_json)
    else:
        path = ls_dirs(path)[n]
        r = get_clf_path(path, verbose = verbose, n = n)
    return r


def get_clf_json(clf_path: str):
    clf_path = get_clf_path(clf_path)
    json_path = os.path.join(clf_path, 'clf.json')
    clf_dict = json_load(json_path)
    return clf_dict


# def get_keys_and_clf(path, verbose = False, n = 0, pmml = False, clf_file = None, fillna=False):
    
#     path = get_clf_path(path, verbose = verbose, n = n)
#     clf_files = os.listdir(path)

#     if 'keys.json' in clf_files:
#         keys = json_load(os.path.join(path, 'keys.json'))
#     elif 'keys_list_pickle' in clf_files:
#         keys = pickle_load(os.path.join(path, 'keys_list_pickle'))
#     else:
#         t_print('ERRROR: could not find keys.')

#     if clf_file in clf_files:
#         clf = pickle_load(os.path.join(path, clf_file))
#     elif 'clf_pickle' in clf_files:
#         clf = pickle_load(os.path.join(path, 'clf_pickle'))
#     elif 'clf_sklearn_pickle' in clf_files:
#         clf = pickle_load(os.path.join(path, 'clf_sklearn_pickle'))
#     else:
#         t_print('ERRROR: could not find clf.')

#     if isinstance(clf, GradientBoostingClassifier) and '_loss' not in dir(clf):
#         print('warning: low version model file and fix _loss.')
#         clf._loss = clf.__dict__['loss_']

#     if pmml:
#         if 'clf.pmml' in clf_files and 'vars_map_main.json' in clf_files:
#             clf_pmml_path = os.path.join(path, 'clf.pmml')
#             clf = Model.fromFile(clf_pmml_path)
#             vars_map_main_path = os.path.join(path, 'vars_map_main.json')
#             vars_map_main = json_load(vars_map_main_path)
#             keys = [k[0] for k in sorted(vars_map_main.items(), key=lambda a: int(a[1][1:]))]
#         else:
#             raise ValueError('could not find clf.pmml and vars_map_main.json')

#     if fillna:
#         if 'fillna.json' in clf_files:
#             fillna_json = json_load(os.path.join(path, 'fillna.json'))
#             r = keys, clf, fillna_json
#         else:
#             t_print('ERRROR: could not find fillna.json')
#     else:
#         r = keys, clf

#     return r


# def get_all_mdl_keys(paths: list):
#     if type(paths) == dict:
#         paths = list(paths.values())
#     elif isinstance(paths, pd.Series):
#         paths = list(paths.values)
#     all_keys = []
#     unkown_paths = []
#     for k, path in enumerate(paths, start=1):
#         t_print(f'\n{k}/{len(paths)}')
#         if os.path.isdir(path):
#             keys = get_keys_and_clf(path, verbose=True)[0]
#             print('len(keys) =', len(keys))
#             all_keys += keys
#         else:
#             unkown_paths.append(path)
#     all_keys = sorted(set(all_keys))
#     t_print('len(all_keys) =', len(all_keys))
#     t_print(f'unkown_paths =', unkown_paths)
#     return all_keys


def find_binary_cols(df: pd.DataFrame):
    binary_cols = []
    not_binary_cols = []
    for col in df:
        sr = df[col]
        sr = sr[~sr.isna()].unique()
        if sorted(sr) == [0, 1]:
            binary_cols.append(col)
        else:
            not_binary_cols.append(col)
    return binary_cols, not_binary_cols


# def cal_y_proba(clf_path: str, x_df: pd.DataFrame, col_name = 'prefix', verbose = False, n = 0, remain = False, pmml = False, fillna = False, binary_mode = False, nan_row_as_nan = True):
    
#     t_print('cal_y_proba()')

#     if not os.path.isdir(clf_path) and clf_path in x_df:
#         t_print(f'cal_y_proba(): info: col = \'{clf_path}\'')
#         y_df = x_df[[clf_path]]
    
#     elif os.path.isdir(clf_path) and clf_path not in x_df:
#         if x_df.isna().any().any():
#             if fillna:
#                 t_print('warning: x_df has NaN.')
#             else:
#                 raise ValueError('x_df has NaN.')
#         else:
#             t_print('NaN not found')
#         clf_path = get_clf_path(clf_path, verbose = verbose, n = n)
#         if fillna:
#             keys, clf, fillna = get_keys_and_clf(clf_path, verbose=True, pmml=pmml, fillna=True)
#         else:
#             keys, clf = get_keys_and_clf(clf_path, verbose=True, pmml=pmml, fillna=False)
#         if col_name in ['prefix', 'exp_name', 'uuid_hex']:
#             clf_json_path = os.path.join(clf_path, 'clf.json')
#             clf_json = json_load(clf_json_path)
#             col_name = clf_json['main'][col_name]
#         t_print('#', col_name)
#         x0_df = x_df.loc[:, keys].astype(float).copy()
#         nan_idx = x0_df.isna().all(axis=1)
#         t_print(1 - x0_df.isna().sum()/len(x0_df))
#         if fillna :
#             x0_df.fillna(fillna, inplace=True)
#         x_np = x0_df.to_numpy()
#         if len(keys) == 1 and remain:
#             y_np = x_np[:, 0]
#         elif not pmml:
#             y_np = clf.predict_proba(x_np)[:, 1]
#         elif pmml:
#             y_np = np.array(clf.predict(x_np))[:, 1]
#         y_df = pd.DataFrame({col_name: y_np}, index=x_df.index)
#         y_df.index.name = 'index'
#         if fillna and nan_row_as_nan:
#             y_df.loc[nan_idx, :] = float('nan')
#         new_nan_idx = y_df.isna().all(axis = 1)
#         t_print('nan_idx.sum() =', nan_idx.sum(), ', coverage =', 1-nan_idx.sum()/len(nan_idx))
#         t_print('new_nan_idx.sum() =', new_nan_idx.sum(), ', coverage =', 1-new_nan_idx.sum()/len(new_nan_idx))
#         if binary_mode and len(keys) > 1 :
#             binary_cols, not_binary_cols = find_binary_cols(x_df[keys])
#             if binary_cols:
#                 binary_idx = (x_df[binary_cols] == 0).all(axis = 1)
#                 not_binary_idx = (x_df[not_binary_cols].isna()).all(axis = 1)
#                 idx = binary_idx & not_binary_idx
#                 y_df.loc[idx, :] = float('nan')
#                 t_print('info: binary_mode = True')
#             else:
#                 t_print('warning: binary col(s) not found when binary_mode = True.')
    
#     y_df = y_df.astype('float64')

#     return y_df


def parse_mdl_str(mdl_str = './mdl_str.txt'):
    if os.path.exists(mdl_str):
        with open(mdl_str, 'r', encoding='utf-8') as f:
            mdl_str = f.read()
    mdl_str = mdl_str.replace('\t', '    ')
    pattern = r'\S+'
    mdl_ls = []
    for mdl_str_k in mdl_str.split('\n'):
        mdl_str_k = mdl_str_k.strip()
        if mdl_str_k == '' or mdl_str_k[0] == '#':
            continue
        ref = re.findall(pattern, mdl_str_k)
        if ref:
            mdl_ls.append(ref)
    mdl_dict = {k[0]: k[1] for k in mdl_ls}
    mdl_sr = pd.Series(mdl_dict)
    mdl_sr.sort_index(inplace=True)
    mdl_sr.name = 'path'
    mdl_sr.index.name = 'mdl_tag'
    return mdl_sr


# def get_FI(path, verbose = False, n = 0, normalized = True):
#     keys, clf = get_keys_and_clf(path, verbose = verbose, n = n)
#     if 'feature_importances_' in dir(clf):
#         if isinstance(clf, LGBMClassifier) or isinstance(clf, XGBClassifier):
#             clf.importance_type = 'gain'
#             if clf.importance_type == 'gain':
#                 # t_print('info: set clf.importance_type = gain')
#                 pass
#             else:
#                 raise ValueError
#         FI_s = pd.Series(dict(zip(keys, clf.feature_importances_)))
#     else:
#         t_print(f"warning: 'feature_importances_' not in {clf.__str__()}")
#         FI_s = pd.Series({k: float('nan') for k in keys})
#     if normalized:
#         FI_s = FI_s / FI_s.sum()
#     return FI_s


# def get_FI_from_exp(exp_path, verbose = False, n = 0, th = .999, how = 'union'):
#     if isinstance(exp_path, pd.Series):
#         exp_path = exp_path.to_list()
#     if how not in ['union', 'intersec']:
#         raise ValueError(f'how = {how}')
#     if type(exp_path) == list:
#         paths = exp_path
#     elif type(exp_path) == str:
#         paths = ls_dirs(exp_path)
#         if len(paths) == 0:
#             paths = [exp_path]
#     top_keys_ls = []
#     k = 0
#     for path in paths:
#         k += 1
#         FI_df = get_FI(path, verbose=verbose, n = n)
#         top_keys = get_FI_top_n(FI_df, th = th)
#         top_keys_ls.append(top_keys)
#         t_print(f'{k}/{len(paths)} top_features: {len(FI_df)} -> {len(top_keys)}')
#     merged_features = set(top_keys_ls[0])
#     for features in top_keys_ls:
#         if how == 'union':
#             merged_features |= set(features)
#         elif how == 'intersec':
#             merged_features &= set(features)
#     merged_features = sorted(merged_features)
#     if merged_features:
#         t_print('len(merged_features) =', len(merged_features))
#     else:
#         raise ValueError(f'merged_features = [] in {exp_path}')
#     return merged_features


# def get_FI_from_clfs(clfs_path, verbose = False):
#     path = parse_clfs(clfs_path).iloc[0, :]['save_path']
#     FI_s = get_FI(path, verbose = verbose)
#     return FI_s


def get_FI_top_n(mean_FI_sr, th = .999):
    if type(mean_FI_sr) == dict:
        mean_FI_sr = pd.Series(mean_FI_sr)
    mean_FI_sr = mean_FI_sr[mean_FI_sr > 0]
    if th == 1:
        keys = mean_FI_sr.index.to_list()
    elif th < 1:
        for k in range(1, len(mean_FI_sr)+1):
            FI_nlargest = mean_FI_sr.nlargest(k)
            if FI_nlargest.sum() >= th:
                break
        keys = FI_nlargest.keys().to_list()
    r = [k for k in mean_FI_sr.index if k in keys]
    return r


# def get_FI_df(features: list, clf: object):
#     if isinstance(clf, LGBMClassifier):
#         clf.importance_type = 'split'
#         split_dict = dict(zip(features, clf.feature_importances_))
#         clf.importance_type = 'gain'
#         gain_dict = dict(zip(features, clf.feature_importances_))
#         FI_df = pd.DataFrame({
#             'split_origin': split_dict,
#             'gain_origin': gain_dict,
#         })
#         FI_df['split_normalized'] = FI_df['split_origin'] / FI_df['split_origin'].sum()
#         FI_df['gain_normalized'] = FI_df['gain_origin'] / FI_df['gain_origin'].sum()
#         FI_df.index.name = 'feature'
#         t_print('FI_df =')
#         t_print(FI_df)
#         FI_df = save_and_reset_index(FI_df)
#     elif isinstance(clf, XGBClassifier):
#         FI_dict = {}
#         importance_type_ls = ['weight', 'gain', 'cover', 'total_gain', 'total_cover']
#         for importance_type in importance_type_ls:
#             clf.importance_type = importance_type
#             FI_dict[importance_type] = clf.feature_importances_
#         FI_df = pd.DataFrame(FI_dict, index=features)
#         FI_df.index.name = 'feature'
#         t_print('FI_df =')
#         t_print(FI_df)
#         FI_df = save_and_reset_index(FI_df)
#         clf.importance_type = None
#     return FI_df


def show_mdl_stat_info(stat_df, abstract_cols = ['len', 'dim', 'validate_ks', 'test_ks', 'overfit', 'decline'], verbose = True):
    # abstract_cols = ['dim', 'validate_ks', 'test_ks', 'overfit', 'decline']
    n = len(stat_df)
    min_s = stat_df.loc[:, abstract_cols].astype(float).min().round(3)
    mean_s = stat_df.loc[:, abstract_cols].astype(float).mean().round(3)
    max_s = stat_df.loc[:, abstract_cols].astype(float).max().round(3)
    info_df = pd.DataFrame({
        'min': min_s, 
        'mean': mean_s, 
        'max': max_s})
    info_df = info_df.T
    info_df['len'] = info_df['len'].astype(int)
    info_df['dim'] = info_df['dim'].astype(int)
    if verbose:
        prefix = stat_df['prefix'].unique().tolist()
        print(f'# prefix = {prefix}, n = {n}', )
        print(info_df)
    return info_df


# def select_mdl(stat_df: pd.DataFrame, dim = float('inf'), validate_ks = .01, test_ks = .01, overfit = .9, verbose = False):
#     idx = (stat_df['overfit'].abs() < overfit) & (stat_df['dim'] <= dim) & (stat_df['validate_ks'] > validate_ks) & (stat_df['test_ks'] > test_ks)
#     show_cols = ['prefix', 'dim', 'save_path', 'validate_ks', 'test_ks', 'overfit', 'decline']
#     show_df = stat_df.loc[idx, show_cols]
#     if verbose:
#         print(show_df)
#     return show_df


# class ClfStat():

#     def __init__(self, df = pd.DataFrame(), path = '') -> None:
#         self.path = path
#         if len(df) > 0:
#             df = df.copy()
#             df['date'] = df['date'].astype('datetime64[ns]')
#             self.df = df
#         elif os.path.isdir(path):
#             self.df = parse_clfs(path)
#         elif os.path.isfile(path):
#             df = auto_reader(path)
#             df['date'] = pd.to_datetime(df['date'])
#             self.df = df
#         else:
#             raise ValueError('No clf_stat.')
#         self.abstract_cols = ['len', 'dim', 'validate_ks', 'test_ks', 'overfit', 'decline']
#         t_print('info: clf_stat()')
#         show_mdl_stat_info(stat_df=self.df, abstract_cols=self.abstract_cols)
#         return None
    
#     def abstract(self):
#         r = show_mdl_stat_info(self.df, abstract_cols=self.abstract_cols)
#         return r 

#     def renew(self):
#         path = self.path
#         if os.path.exists(path):
#             self.df = parse_clfs(path)
#             # self.abstract = show_mdl_stat_info(df)
#             # self.manual_select_dict = {}
#         else:
#             print(f'renew(): warning: path = \'{path}\' does not exits.')
#         return None


#     def manual_select(
#         self, 
#         dim = float('inf'), date = 'now', 
#         validate_ks = 'mean', test_ks = 'mean', 
#         overfit = .9, decline = float('inf'), 
#         th = .9,
#         sigma = 2,
#         len_sample = 1,
#         ret = 'df'):
        
#         df = self.df

#         if date == 'now':
#             date = datetime.datetime.now()
#         df = df.loc[df['date'] <= date, :]

#         if sigma > 0:
#             validate_idx = (df['validate_ks'] - df['validate_ks'].mean()).abs() < sigma * df['validate_ks'].std()
#             test_idx = (df['test_ks'] - df['test_ks'].mean()).abs() < sigma * df['test_ks'].std()
#             sigma_idx = validate_idx & test_idx
#             df = df.loc[sigma_idx, :]

#         if dim == 'mean':
#             dim = df['dim'].mean()
#         elif dim == 'min':
#             dim = df['dim'].min()
#         elif dim == 'max':
#             dim = df['dim'].max()
#         elif isinstance(dim, str) and 'min' in dim:
#             dim = sorted(df['dim'].unique())[int(dim[-1])-1]
#         df = df.loc[df['dim'] <= dim, :]

#         if overfit == 'mean':
#             overfit = df['overfit'].abs().mean()
#         if overfit == 'max':
#             overfit = df['overfit'].abs().max()
#         elif type(overfit) == list:
#             overfit = df['overfit'][df['overfit'].abs() < overfit[1]].abs().mean()
#         elif overfit == 'auto':
#             overfit_min = df['overfit'].abs().min()
#             overfit_mean = df['overfit'].abs().mean()
#             overfit = (overfit_min+overfit_mean)/2
#         df = df.loc[df['overfit'].abs() <= overfit, :]

#         if validate_ks == 'mean':
#             validate_ks = df['validate_ks'].mean()
#         elif type(validate_ks) == str and '+' in validate_ks:
#             k = float(validate_ks.split('+')[1])
#             validate_ks = df['validate_ks'].mean() + k*df['validate_ks'].std()

#         if test_ks == 'mean':
#             test_ks = df['test_ks'].mean()
#         elif type(test_ks) == str and '+' in test_ks:
#             k = float(test_ks.split('+')[1])
#             test_ks = df['test_ks'].mean() + k*df['test_ks'].std()

#         if decline == 'mean':
#             decline = df['decline'].abs().mean()

#         idx = True
#         idx = idx & (df['validate_ks'] >= validate_ks)
#         idx = idx & (df['test_ks'] >= test_ks)
#         idx = idx & (df['decline'].abs() <= decline)
#         idx = idx & (df['len'] >= len_sample)

#         select_df = df.loc[idx, :]
#         if len(select_df) == 0:
#             t_print(f'\nError when overfit = {overfit:.3f}, dim = {dim}\n')
#             count_down(16)
#             raise ValueError('len(select_df) == 0')
#         else:
#             print('head0:\n')
#             print('    '.join([str(k) for k in select_df[['prefix', 'save_path', 'validate_ks', 'test_ks', 'overfit']].iloc[0, :]]))
#             print()
            
#         print('info: from manual_select()')
#         abstract_df = show_mdl_stat_info(stat_df=select_df, abstract_cols=self.abstract_cols)

#         if ret == 'df':
#             r = select_df

#         elif ret == 'keys':
#             paths = select_df['save_path'].to_list()
#             valid_keys = []
#             t_print(f'get valid_cols in {len(paths)} clfs.')
#             for p in paths:
#                 s = get_FI(p)
#                 s = get_FI_top_n(s, th=th)
#                 valid_keys += s
#             valid_keys = sorted(set(valid_keys))
#             t_print(f'len valid_cols = {len(valid_keys)}, th = {th}.')
#             r = valid_keys

#         elif ret == 'abstract':
#             r = abstract_df

#         elif ret == 'select':
#             score_df = pd.DataFrame(index=select_df.index)
#             score_df['overfit_s'] = (select_df['overfit'].abs() / .05).apply(int)
#             score_df['dim_s'] = (select_df['dim'] / 20).apply(int)
#             score_df['score'] = (select_df['validate_ks']+select_df['test_ks'])/2
#             score_df.sort_values(by='score', ascending=False, inplace=True)
#             dup = score_df.loc[:, ['overfit_s', 'dim_s']].duplicated()
#             score_df = score_df.loc[~dup, :]
#             r = select_df.loc[score_df.index, :]

#         return r

#     def plot(self, stat_df = pd.DataFrame(), dots = [], title = '', alpha_min = .2, figsize = [8, 4.5], fontsize = 12):
#         if len(stat_df) == 0:
#             stat_df = self.df
#         if title == '':
#             title = ','.join(sorted(stat_df['prefix'].unique()))
#         plot_mdl_stat_df(stat_df=stat_df, title=title, dots=dots, alpha_min = alpha_min, figsize = figsize, fontsize=fontsize)
    
#     None


# clf_stat = ClfStat


# def plot_mdl_stat_df(stat_df, title = '', dots = [], alpha_min = .2, figsize = [8, 4.5], fontsize = 12):

#     xs = stat_df['validate_ks']
#     ys = stat_df['test_ks']
#     cs = stat_df['overfit']

#     x_m = xs.mean()
#     y_m = ys.mean()
#     x_s = xs.std()
#     y_s = xs.std()

#     plt.figure(figsize=figsize)

#     plt.plot([x_m], [y_m], 'C7o', alpha = .5)
#     plt.axhline(y=y_m+y_s, color = 'C7', linestyle = '--', alpha = .5)
#     plt.axhline(y=y_m-y_s, color = 'C7', linestyle = '--', alpha = .5)
#     plt.axvline(x=x_m+x_s, color = 'C7', linestyle = '--', alpha = .5)
#     plt.axvline(x=x_m-x_s, color = 'C7', linestyle = '--', alpha = .5)

#     # plt.axis([
#     #     xs.min(), xs.max(),
#     #     ys.min(), ys.max(),
#     #     ])

#     overfit_to_color_n = {
#         v: k for k, v in enumerate(sorted(cs.abs().unique()))
#         }
#     color_ls = plt.cm.RdYlGn(np.linspace(0, 1, len(overfit_to_color_n)))
#     color_ls = color_ls[::-1]

#     for xk, yk, ck in zip(xs, ys, cs):
#         color = color_ls[overfit_to_color_n[abs(ck)]]
#         alpha = (cs.abs() >= abs(ck)).sum() / len(cs)
#         plt.plot([xk], [yk], '.', color = color, alpha = max([(1-alpha)/2, alpha_min]))
#         plt.text(xk, yk, f'{100*ck:.0f}', 
#         horizontalalignment = 'center',
#         verticalalignment = 'center',
#         color = color,
#         alpha = max([alpha, alpha_min]),
#         fontsize=fontsize
#         )
#     for xk, yk in dots:
#         plt.plot([xk], [yk], 'C7^', alpha = .8)
#         plt.text(xk, yk, f'[{xk:.2f}, {yk:.2f}]', 
#         color = 'C7', alpha = .8, 
#         verticalalignment = 'top')
#     plt.grid(which='both', axis='both')
#     plt.xlabel('validate_ks')
#     plt.ylabel('test_ks')
#     plt.title(f'{title} (n={len(xs)})')
#     plt.show()

#     return None


# def add_mdl_files(path, hash_files = False, timestamp = None):

#     clf_path = get_clf_path(path)

#     try:
#         save_FI_csv(clf_path)
#         t_print('FI.csv saved.')
#     except:
#         print('ERROR: save_FI_csv(path)')

#     try:
#         save_clf_pmml(clf_path, timestamp = timestamp)
#         t_print('clf.pmml saved.')
#     except Exception as e:
#         print(f'ERROR: save_clf_pmml(): {e}')

#     try:
#         save_pmml_keys(clf_path)
#         t_print('clf key files saved.')
#     except:
#         print('ERROR: save_pmml_keys(path)')

#     if hash_files:
#         dump_files_hash(clf_path)
    
#     return None


def color_results_csv(df: pd.DataFrame, xlsx_path = None):
    if type(df) == str:
        csv_path = df
        df = auto_reader(csv_path)
        if xlsx_path is None:
            xlsx_path = csv_path[:-4] + '.xlsx'
    else:
        if type(xlsx_path) != str:
            raise ValueError(f'illegal xlsx_path = {xlsx_path}')
    cful = colorful(df)
    color_cols = [
        'dim',
        'validate_auc', 
        'test_auc', 
        'validate_ks',
        'test_ks',
        'overfit',
        # 'decline'
        ]
    cful.set_conditional_formatting(cols=color_cols)
    format_cols = [
        'train_auc',
        'validate_auc', 
        'test_auc', 
        'train_ks',
        'validate_ks',
        'test_ks',
        'overfit',
        'decline',
        'ks_std',
        ]
    cful.set_number_format(cols=format_cols, format='0.0%')
    cful.set_font(df['choose'][df['choose']].index)
    cful.save(xlsx_path)
    t_print(xlsx_path)
    return None



#%%


# def fix_pmml_timestamp(path, timestamp):
#     if isinstance(timestamp, str) and len(timestamp) == 17:
#         timestamp = datetime.datetime.strptime(timestamp, '%Y%m%d%a%H%M%S')
#     timestamp_str = timestamp.strftime('%Y-%m-%dT%H:%M:%S+08:00')
#     pmml_str = read_txt(path)
#     Timestamp_pattern = r'<Timestamp>.+</Timestamp>'
#     pmml_str = re.sub(Timestamp_pattern, f'<Timestamp>{timestamp_str}</Timestamp>', pmml_str)
#     save_txt(pmml_str, path)
#     return None


# def dump_clf_pmml(clf, path, with_repr=True, timestamp = None):
#     pipeline = PMMLPipeline([
#         ("classifier", clf)
#     ])
#     sklearn2pmml(pipeline, path, with_repr=with_repr)
#     if timestamp is not None:
#         fix_pmml_timestamp(path, timestamp)
#     return None


# def save_FI_csv(clf_path: str, exist='skip') -> pd.DataFrame:
#     clf_path = get_clf_path(clf_path)
#     FI_sr = get_FI(clf_path)
#     FI_df = pd.DataFrame({'importances': FI_sr})
#     FI_df.index.name = 'feature'
#     FI_df.reset_index(inplace=True)
#     FI_df.index.name = 'index'
#     FI_df.index += 1
#     FI_csv_path = os.path.join(clf_path, 'FI.csv')
#     if os.path.exists(FI_csv_path) and exist == 'skip':
#         print('FI.csv exists.')
#     else:
#         FI_df.to_csv(FI_csv_path)
#     return FI_df


# def save_clf_pmml(clf_path: str, exist='skip', timestamp = None) -> None:
#     clf_path = get_clf_path(clf_path)
#     keys, clf = get_keys_and_clf(clf_path)
#     vars_map_all = {key: f'x{k+1}' for k, key in enumerate(keys)}
#     clf_pmml_path = os.path.join(clf_path, 'clf.pmml')
#     if os.path.exists(clf_pmml_path) and exist == 'skip':
#         print('clf.pmml exists.')
#     else:
#         dump_clf_pmml(clf, clf_pmml_path, timestamp = timestamp)
#     vars_map_all_path = os.path.join(clf_path, 'vars_map_all.json')
#     if os.path.exists(vars_map_all_path) and exist == 'skip':
#         print('vars_map_all.json exists.')
#     else:
#         json_dump(vars_map_all, vars_map_all_path)
#     return None


# def save_pmml_keys(clf_path, exist='skip') -> None:
#     clf_path = get_clf_path(clf_path)
#     clf_pmml_path = os.path.join(clf_path, 'clf.pmml')
#     model = Model.fromFile(clf_pmml_path)
#     main_vars = model.inputNames
#     vars_map_all_path = os.path.join(clf_path, 'vars_map_all.json')
#     vars_map_all = json_load(vars_map_all_path)
#     vars_map_main = {k: v for k, v in vars_map_all.items() if v in main_vars}
#     vars_map_main_path = os.path.join(clf_path, 'vars_map_main.json')
#     if os.path.exists(vars_map_main_path) and exist == 'skip':
#         print('vars_map_main.json exists.')
#     else:
#         json_dump(vars_map_main, vars_map_main_path)
#     return None


# def save_files_sha256(clf_path: str):
#     clf_path = get_clf_path(clf_path)
#     files_sha256 = get_files_sha256(clf_path)
#     files_sha256_path = os.path.join(clf_path,'files_sha256.json')
#     if 'files_sha256.json' in files_sha256:
#         del files_sha256['files_sha256.json']
#     json_dump(files_sha256, files_sha256_path)
#     return None

def save_files_sha256(path: str):
    paths = ls_files(path)
    d = {
        'file': [os.path.basename(k) for k in paths],
        'sha256': [get_file_hash(k) for k in paths],
        }
    if 'file_sha256.csv' in d:
        del d['file_sha256.csv']
    df = pd.DataFrame(d)
    df.index += 1
    df.index.name = 'index'
    save_path = os.path.join(path, 'file_sha256.csv')
    df.to_csv(save_path, sep='\t', index=None)
    return df


# def sample_rows(df, n, stratify = None):
#     y = [0] * len(df)
#     if type(stratify) == list:
#         stratify = df.loc[:, stratify]
#     sample_df = train_test_split(
#         df, y, 
#         test_size=n, 
#         stratify=stratify,
#         random_state=0
#         )[1]
#     sample_df.sort_index(inplace = True)
#     return sample_df


def sample_prod(a: int, b: float):
    if type(b) == int and type(a) == float:
        b, a = a, b
    if (a <= 1) or (not (0 < b < 1)):
        raise ValueError('a <= 1 or b not in (0, 1)')
    r = round( a * b )
    if r == 0:
        r = 1
    elif r == a:
        r = a - 1
    return r


def sample_rows_0(df: pd.DataFrame, a = .2):
    if a < 1:
        n = sample_prod(len(df), a)
    else:
        n = round(a)
    r = df.sample(n, random_state=0)
    return r


def sample_rows(df: pd.DataFrame, a, stratify = []):

    if len(df) < 2:
        raise ValueError('len(df) < 2')

    if .5 < a < 1 :
        a = 1 - a
        reverse = True
    elif a > len(df)/2:
        a = len(df) - a
        reverse = True
    else:
        reverse = False
        
    if stratify == []:
        if a < 1:
            n = sample_prod(len(df), a)
        else:
            n = a
        r = df.sample(n, random_state=0)
    else:
        if a > 1:
            a = a / len(df)
        r = df.groupby(by=stratify).apply(functools.partial(sample_rows_0, a = a))
        r.index = [k[-1] for k in r.index]
        r.sort_index(inplace=True)

    if reverse:
        r = df.loc[~df.index.isin(r.index), :]

    return r


# get_vis = lambda s: ''.join([k for k in s if k.isprintable()]) if type(s) == str else s

def get_vis(s: str):
    if isinstance(s, str):
        s = ''.join([k for k in s if k.isprintable()])
    return s

is_invis = lambda s: not all([k.isprintable() for k in s ])

def drop_invisable(df: pd.DataFrame):
    df.index = df.index.to_series().apply(get_vis)
    df.index.name = get_vis(df.index.name)
    df.columns = df.columns.to_series().apply(get_vis)
    cols = df.select_dtypes(exclude='number').columns.to_list()
    for k, col in enumerate(cols, start=1):
        # df[col] = df[col].apply(get_vis)
        t_print(f'{k}/{len(cols)}', col)
        df[col] = series_multiprocessing_apply(df[col], get_vis)
    t_print('done')
    return df

#%%
# openpyxl
try:
    from openpyxl import load_workbook
    from openpyxl.formatting.rule import ColorScaleRule
    from openpyxl.styles import Color, Alignment, Font
    from openpyxl.utils import get_column_letter

    bold_font = Font(bold=True)
    color_scale_rule = ColorScaleRule(
        start_value=0,
        start_type='percentile',
        start_color=Color(rgb='63BE7B'),  # green
        
        mid_value=50,
        mid_type='percentile',
        mid_color=Color(rgb='FFEB84'), # yellow

        end_value=100,
        end_type='percentile',
        end_color=Color(rgb='F8696B')    # red
    )

except:
    print('No pkg: openpyxl')


def dataframe_to_workbook(df: pd.DataFrame, sheet_name = 'Sheet1'):
    with io.BytesIO() as buffer:
        with pd.ExcelWriter(buffer, engine='openpyxl') as writer:
            df.to_excel(writer, sheet_name=sheet_name)
        buffer.seek(0)
        workbook = load_workbook(buffer)
    return workbook


def cols_to_letters(df: pd.DataFrame, cols = []):
    letters = []
    for col in cols:
        if type(col) == str:
            col = df.columns.get_loc(col)
        if type(col) == int:
            lt = get_column_letter(col+2)
        letters.append(lt)
    return letters

def index_to_loc(df: pd.DataFrame, index = []):
    if index is None:
        index = df.index
    s = [df.index.get_loc(k)+2 for k in index]
    return s

def cols_to_loc(df: pd.DataFrame, cols = []):
    if cols is None:
        cols = df.columns
    s = [df.columns.get_loc(k)+2 for k in cols]
    return s


def save_dfs_to_excel(xlsx_dict: dict, path: str, index = True, makedirs = True):
    if isinstance(xlsx_dict, list):
        xlsx_dict = {f'sheet_{k}': v for k, v in enumerate(xlsx_dict)}
    if not os.path.isdir(os.path.dirname(path)) and makedirs:
        os.makedirs(os.path.dirname(path), exist_ok=True)
    with pd.ExcelWriter(path) as writer:
        for sheet_name, df in xlsx_dict.items():
            df.to_excel(writer, sheet_name=sheet_name, index=index)
    path_str = colorize_str(path, color=[4, 94])
    print(f'xlsx file saved successfully: {path_str}')
    return None

save_dataframes_to_excel = save_dfs_to_excel


try: 
    class colorful():

        def fix_origin_alignment(self) -> None:
            if self.df.index.dtype.name in ['Int64', 'int64', 'Int32', 'int32', 'float64']:
                horizontal = 'right'
            else:
                horizontal = 'left'
            for k in range(1, self.n_row+1):
                cell = self.ws.cell(row=k, column=1)
                cell.alignment = Alignment(horizontal=horizontal, vertical='center')
            for k in range(1, self.n_col+1):
                cell = self.ws.cell(row=1, column=k)
                cell.alignment = Alignment(horizontal='left', vertical='center')
            return None

        def __init__(self, df: pd.DataFrame, sheet_name  = 'Sheet1' ) -> None:
            if isinstance(df, pd.DataFrame):
                self.df = df
            else:
                path = df
                df = auto_reader(path)
                self.df = df
            self._sheet_name = sheet_name
            self.wb = dataframe_to_workbook(df, sheet_name=sheet_name)
            self.ws = self.wb[sheet_name]
            self.ws.freeze_panes = 'A2'
            self.n_row = self.ws.max_row
            self.n_col = self.ws.max_column
            self.fix_origin_alignment()
            end_col_letter = get_column_letter(self.n_col)
            self.ws.auto_filter.ref = f"A1:{end_col_letter}{self.n_row}"
            return None

        def set_conditional_formatting(self, cols = []) -> None:
            n_row = self.n_row
            cols_in_letter = cols_to_letters(df=self.df, cols=cols)
            for col in cols_in_letter:
                self.ws.conditional_formatting.add(f'{col}2:{col}{n_row}', color_scale_rule)
            return None

        def set_number_format(self, index = None, cols = None, format = '0.000') -> None:
            cols_loc = cols_to_loc(df=self.df, cols=cols)
            index_loc = index_to_loc(df=self.df, index=index)
            for col in cols_loc:
                for row in index_loc:
                    cell = self.ws.cell(row=row, column=col)
                    cell.number_format = format # '0.0%'
            return None

        def set_font(self, index = None, cols = None, format = bold_font):
            cols_loc = cols_to_loc(df=self.df, cols=cols)
            index_loc = index_to_loc(df=self.df, index=index)
            for n in index_loc:
                for k in cols_loc:
                    cell = self.ws.cell(row=n, column=k)
                    cell.font = format
            return None

        def save(self, save_path = ''):
            if save_path == '':
                timestamp = get_timestamp()
                save_path = f'{timestamp}.xlsx'
            elif not save_path.endswith('.xlsx'):
                save_path = f'{save_path}.xlsx'
            self.wb.save(save_path)
            print(f'save_path = \'{save_path}\'')
            return None

        def __str__(self) -> str:
            return str(self.df)

        def __repr__(self) -> str:
            return str(self.df)
except:
    print('No class: colorful')

    # end of class

#%%
# re-write

def rewrited_stable_cumsum(arr, axis=None, rtol=1e-05, atol=1e-08):
    out = np.cumsum(arr, axis=axis, dtype=np.float64)
    return out


def rewrited_binary_clf_curve(y_true, y_score, pos_label=None, sample_weight=None):
    pos_label = 1
    if isinstance(y_true, pd.Series):
        y_true = y_true.to_numpy()
    if isinstance(y_score, pd.Series):
        y_score = y_score.to_numpy()
    y_true = y_true == pos_label
    desc_score_indices = np.argsort(y_score, kind="mergesort")[::-1]
    y_score = y_score[desc_score_indices]
    y_true = y_true[desc_score_indices]
    if sample_weight is not None:
        weight = sample_weight[desc_score_indices]
    else:
        weight = 1.0
    distinct_value_indices = np.where(np.diff(y_score))[0]
    threshold_idxs = np.r_[distinct_value_indices, y_true.size - 1]
    tps = rewrited_stable_cumsum(y_true * weight)[threshold_idxs]
    if sample_weight is not None:
        fps = rewrited_stable_cumsum((1 - y_true) * weight)[threshold_idxs]
    else:
        fps = 1 + threshold_idxs - tps
    return fps, tps, y_score[threshold_idxs]


def rewrited_roc_curve(y_true, y_score, pos_label=None, sample_weight=None, drop_intermediate=True):
    fps, tps, thresholds = rewrited_binary_clf_curve(
        y_true, y_score, pos_label=pos_label, sample_weight=sample_weight
    )
    if drop_intermediate and len(fps) > 2:
        optimal_idxs = np.where(
            np.r_[True, np.logical_or(np.diff(fps, 2), np.diff(tps, 2)), True]
        )[0]
        fps = fps[optimal_idxs]
        tps = tps[optimal_idxs]
        thresholds = thresholds[optimal_idxs]
    tps = np.r_[0, tps]
    fps = np.r_[0, fps]
    thresholds = np.r_[thresholds[0] + 1, thresholds]
    fpr = fps / fps[-1]
    tpr = tps / tps[-1]
    return fpr, tpr, thresholds


def rewrited_auc(x, y):
    direction = 1
    dx = np.diff(x)
    if np.any(dx < 0):
        if np.all(dx <= 0):
            direction = -1
    area = direction * np.trapz(y, x)
    if isinstance(area, np.memmap):
        area = area.dtype.type(area)
    return area


def rewrited_binary_roc_auc_score(y_true, y_score, sample_weight=None):
    if isinstance(y_true, pd.Series):
        y_true = y_true.to_numpy()
    if isinstance(y_score, pd.Series):
        y_score = y_score.to_numpy()
    fpr, tpr, _ = rewrited_roc_curve(y_true, y_score, sample_weight=sample_weight)
    ret = rewrited_auc(fpr, tpr)
    ret = float(ret)
    return ret

roc_curve = rewrited_roc_curve
roc_auc_score = rewrited_binary_roc_auc_score

# from sklearn.metrics import roc_auc_score, roc_curve, accuracy_score, precision_score, recall_score

#%%
# others





